#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import os.path as op
import pandas as pd

import numpy as np
from scipy import interpolate
import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1 import make_axes_locatable


# TODO: updated plot library, temporal
# new develops at swan/plots/ common.py stat.py nonstat.py

def aux_quiver(X, Y, var, vdir):
    '''
    interpolates var and plots quiver with var_dir. Requires open figure

    var  - variable module
    vdir - variable direction (º clockwise relative to North)
    '''

    size = 30  # quiver mesh size

    # var and dir interpolators 
    vdir_f = vdir.copy()
    vdir_f[np.isnan(vdir_f)] = 0
    f_dir = interpolate.interp2d(X, Y, vdir_f, kind='linear')

    var_f = var.copy()
    var_f[np.isnan(var_f)] = 0
    f_var = interpolate.interp2d(X, Y, var_f, kind='linear')

    # generate quiver mesh
    x_q = np.linspace(X[0], X[-1], num = size)
    y_q = np.linspace(Y[0], Y[-1], num = size)

    # interpolate data to quiver mesh
    vdir_q = f_dir(x_q, y_q)
    var_q = f_var(x_q, y_q)

    # u and v dir components
    u = np.sin(np.deg2rad(vdir_q))
    v = np.cos(np.deg2rad(vdir_q))

    # plot quiver
    plt.quiver(x_q, y_q, -u*var_q, -v*var_q) #, width=0.003, scale=1, scale_units='inches')

def plot_var_times(xds_out_case, var_name, p_export_case, quiver=False,
                   np_shore=np.array([]), cmap='jet'):
    '''
    Plots non-stationary SWAN execution output for selected var and case

    xds_out_case    - swan output (xarray.Dataset)
    var_name        - 'Hsig', 'Tm02', 'Tpsmoo'
    p_export        - path for exporting figures

    opt. args
    quiver     - True for adding directional quiver plot
    np_shore   - shoreline, np.array x = np_shore[:,0] y = np.shore[:,1]
    cmap       - matplotlib colormap
    '''

    if not op.isdir(p_export_case): os.makedirs(p_export_case)

    # iterate case output over time
    for t in xds_out_case.time.values[:]:
        xds_oct = xds_out_case.sel(time=t)

        # time string
        t_str = pd.to_datetime(str(t)).strftime('%Y%m%d-%H%M')

        # get mesh data from output dataset
        X = xds_oct.X.values[:]
        Y = xds_oct.Y.values[:]

        # get variable and units
        var = xds_oct[var_name].values[:]
        var_units = xds_oct[var_name].attrs['units']

        # new figure
        fig, ax0 = plt.subplots(nrows=1, figsize=(12, 12))
        var_title = '{0}'.format(var_name)  # title

        # pcolormesh
        ocean = plt.get_cmap('jet')  # colormap
        im = plt.pcolormesh(X, Y, var, cmap=cmap)

        # add quiver plot 
        if quiver:
            var_title += '-Dir'
            var_dir = xds_oct.Dir.values[:]
            aux_quiver(X, Y, var, var_dir)

        # shoreline
        if np_shore.any():
            x_shore = np_shore[:,0]
            y_shore = np_shore[:,1]
            plt.plot(x_shore, y_shore,'.', color='dimgray', markersize=3)

        # customize pcolormesh
        plt.title('{0} (t={1})'.format(var_title, t_str),
                  fontsize = 12, fontweight='bold')
        plt.xlabel(xds_oct.attrs['xlabel'], fontsize = 12)
        plt.ylabel(xds_oct.attrs['ylabel'], fontsize = 12)

        plt.axis('scaled')
        plt.xlim(X[0], X[-1])
        plt.ylim(Y[0], Y[-1])

        # add custom colorbar
        divider = make_axes_locatable(ax0)
        cax = divider.append_axes("right", size="5%", pad=0.1)
        plt.colorbar(im, cax=cax)
        plt.ylabel('{0} ({1})'.format(var_name, var_units), fontsize = 12)

        # export fig
        p_ex = op.join(p_export_case, 'outmap_{0}_{1}.png'.format(var_name, t_str))
        fig.savefig(p_ex)

        # close fig 
        plt.close()

def plot_output_nonstat(xds_out, var_name, p_export, quiver=False,
                        np_shore=np.array([])):
    '''
    Plots non-stationary SWAN execution output for selected var, for every case

    xds_out    - swan output (xarray.Dataset)
    var_name   - 'Hsig', 'Tm02', 'Tpsmoo'
    p_export   - path for exporting figures

    opt. args
    quiver     - True for adding directional quiver plot
    np_shore   - shoreline, np.array x = np_shore[:,0] y = np.shore[:,1]
    '''

    # make export dir
    if not op.isdir(p_export): os.makedirs(p_export)

    for case_ix in xds_out.case.values[:]:

        # select case
        xds_out_case = xds_out.sel(case=case_ix)

        # output case subfolder
        case_id = '{0:04d}'.format(case_ix)
        p_export_case = op.join(p_export, case_id)

        # plot variable times
        plot_var_times(
            xds_out_case, var_name, p_export_case,
            quiver=quiver, np_shore=np_shore)

def plot_points_times(xds_out_case, p_export_case):
    '''
    Plots non-stationary SWAN points output for selected case

    xds_out_case   - swan case output (xarray.Dataset)
    p_export_case  - path for exporting figures
    '''

    if not op.isdir(p_export_case): os.makedirs(p_export_case)

    # iterate over points
    n_pts = len(xds_out_case.point)
    for i in range(n_pts):

        # get point variables
        xds_pt = xds_out_case.isel(point=i)

        hs = xds_pt.HS.values[:]
        tm = xds_pt.TM02.values[:]
        tp = xds_pt.RTP.values[:]
        dr = xds_pt.DIR.values[:]

        # plot and save figure series of each output point
        fig, (ax1, ax2, ax3, ax4) = plt.subplots(4, 1, figsize=(12, 12))

        ax1.plot(hs, '.', color = 'b', markersize=2, label="Hs [m]")
        #ax1.set_xlim([time[0][0], time[0][-1]])
        plt.setp(ax1.get_xticklabels(), visible=False)
        ax1.set_title('Significant Wave Height [m]', fontweight = 'bold')

        ax2.plot(tm, '.', color = 'b', markersize=2)
        #ax3.set_xlim([time[0][0], time[0][-1]])
        plt.setp(ax2.get_xticklabels(), visible=False)
        ax2.set_title('Mean Period [s]', fontweight = 'bold')

        ax3.plot(tp, '.', color = 'b', markersize=2)
        #ax3.set_xlim([time[0][0], time[0][-1]])
        plt.setp(ax3.get_xticklabels(), visible=False)
        ax3.set_title('Peak Period [s]', fontweight = 'bold')

        ax4.plot(dr, '.', color = 'b', markersize=2)
        ax4.set_ylim([0, 360])
        #ax4.set_xlim([time[0][0], time[0][-1]])
        plt.setp(ax4.get_xticklabels(), visible=True)
        ax4.set_title('Wave direction [º]', fontweight = 'bold')

        # export fig
        p_ex = op.join(p_export_case, 'point_{0}.png'.format(i))
        fig.savefig(p_ex)

        # close fig 
        plt.close()

def plot_output_points(xds_out, p_export):
    '''
    Plots SWAN execution output table points time series

    xds_out    - swan points output (xarray.Dataset)
    p_export   - path for exporting figures
    '''

    # make export dir
    if not op.isdir(p_export): os.makedirs(p_export)

    for case_ix in xds_out.case.values[:]:

        # select case
        xds_out_case = xds_out.sel(case=case_ix)

        # output case subfolder
        case_id = '{0:04d}'.format(case_ix)
        p_export_case = op.join(p_export, case_id)

        # plot variable times
        plot_points_times(xds_out_case, p_export_case)

def plot_storm_track(lon0, lon1, lat0, lat1, pd_storm, p_export, np_shore=np.array([])):
    '''
    Plots SWAN execution output table points time series

    lon0, lon1  - longitude axes limits (lon0, lon1)
    lat0, lat1  - latitude axes limits (lat0, lat1)
    pd_storm    - storm track pandas.DataFrame (x0, y0, R as metadata)
    p_export    - path for exporting figure

    opt. args
    np_shore   - shoreline, np.array x = np_shore[:,0] y = np.shore[:,1]
    '''

    # make export dir
    if not op.isdir(p_export): os.makedirs(p_export)

    # get storm track data
    xt = pd_storm.lon
    yt = pd_storm.lat
    pmin = pd_storm.p0[0]
    vmean = pd_storm.vf[0]

    # get storm metadata
    x0 = pd_storm.x0
    y0 = pd_storm.y0
    R = pd_storm.R

    # circle angles
    ang = tuple(np.arange(0, 2*np.pi, 2*np.pi/1000))

    # circle coordinates
    x = R * np.cos(ang) + x0
    y = R * np.sin(ang) + y0

    # plot and save figure
    fig = plt.figure(figsize=(12, 12))

    # plot track
    plt.plot(xt, yt, 'o-', linewidth=2, color='purple', label='Great Circle')

    # plot small circle and center
    plt.plot(x, y, '-', linewidth=2, color='green', label='')
    plt.plot(x0, y0, '.', markersize=10, color='dodgerblue', label='')

    # plot shoreline
    if np_shore.any():
        x_shore = np_shore[:,0]
        y_shore = np_shore[:,1]
        plt.plot(x_shore, y_shore,'.', color='dimgray', markersize=3, label='')

    # plot parameters
    plt.axis('scaled')
    plt.xlim(lon0, lon1)
    plt.ylim(lat0, lat1)
    plt.xlabel('Longitude')
    plt.ylabel('Latitude')
    plt.title('Pmin: {0} hPa  /  V: {1} km/h'.format(pmin, vmean))
    plt.legend()

    # export fig
    p_save = op.join(p_export, 'track_coords.png')
    fig.savefig(p_save)

    # close fig
    plt.close()




#!/usr/bin/env python
# -*- coding: utf-8 -*-

# common
import os
import os.path as op
import itertools
import calendar
from datetime import datetime

# pip
import numpy as np
import xarray as xr
import matplotlib.pyplot as plt

from scipy.stats import gaussian_kde, probplot
from sklearn.metrics import mean_squared_error


# import constants
from .config import _faspect, _fsize, _fdpi

def rmse(predictions, targets):
    return np.sqrt(((predictions - targets) ** 2).mean())

def bias(predictions, targets):
    return sum(predictions-targets)/len(predictions)

def si(predictions, targets):
    S = predictions.mean()
    O = targets.mean()
    return np.sqrt(sum(((predictions-S)-(targets-O))**2)/((sum(targets**2))))


def scatter_QQ(xds_1, xds_2, vn, xlabel='', ylabel=''):
    'Scatter plot v_1 - v_2 and adds rmse, si, bias and QQ plot'

    cmap='jet'

    # data
    x = xds_1[vn]
    y = xds_2[vn]

    # figure
    fig, ax = plt.subplots(1,1, constrained_layout=True,
                           figsize=(_faspect*_fsize,_faspect*_fsize))

    # stack data and gaussian kde
    xy = np.vstack([x, y])
    xy = xy[~np.isnan(xy).any(axis=1)]  # remove nans

    z = gaussian_kde(xy)(xy)

    # sort data
    idx = z.argsort()
    x2, y2, z = x[idx], y[idx], z[idx]

    ax.scatter(x2, y2, c=z, s=0.6, cmap=cmap)
    ax.set_xlabel('{0} {1}'.format(vn, xlabel))
    ax.set_ylabel('{0} {1}'.format(vn, ylabel))

    xmax = max(x) + 0.1
    ymax = max(y) + 0.1
    maxt = np.ceil(max(xmax, ymax))
    ax.set_xlim(0, maxt)
    ax.set_ylim(0, maxt)

    ax.plot([0, maxt],[0, maxt],'-k', linewidth=0.6)
    ax.set_xticks(np.linspace(0, maxt, 5))
    ax.set_yticks(np.linspace(0, maxt, 5))
    ax.set_aspect('equal')

    xqq = probplot(x2, dist="norm")
    yqq = probplot(y2, dist="norm")
    ax.plot(xqq[0][1], yqq[0][1], "o", markersize=1, color='gold' ,label='Q-Q plot')

    mse = mean_squared_error(x2, y2)
    rmse_e = rmse(x2, y2)
    BIAS = bias(x2, y2)
    SI = si(x2, y2)
    label = '\n'.join((
             r'RMSE = %.2f' % (rmse_e, ),
             r'mse =  %.2f' % (mse,  ),
             r'BIAS = %.2f' % (BIAS,  ),
             R'SI = %.2f' % (SI,  )))
    ax.text(0.05, 0.8, label, transform=ax.transAxes)
    ax.legend(loc='upper right')

def compare_series(xds_1, xds_2, vns, label_1='', label_2=''):

    # one plot for each variable at vns
    n_plots = len(vns)

    # figure
    fig, axs = plt.subplots(
        n_plots, 1, sharex=True,
        figsize=(_faspect*_fsize, _fsize)
    )

    # plot variables
    for c, vn in enumerate(vns):
        axs[c].scatter(xds_1['time'], xds_1[vn], label=label_1, color='r',
                    marker='.', s=6)
        axs[c].scatter(xds_2['time'], xds_2[vn],  label=label_2, color='k',
                    marker='.', s=6)
        axs[c].set_ylabel(vn)

    # config first plot
    axs[0].set_xlim(xds_1['time'][0], xds_1['time'][-1])
    axs[0].legend()




import os
import os.path as op
from typing import List

import numpy as np
import pandas as pd
from hydromt_sfincs import SfincsModel

from .._base_wrappers import BaseModelWrapper


class SfincsModelWrapper(BaseModelWrapper):
    """
    Wrapper for the SFINCS model.

    Attributes
    ----------
    default_parameters : dict
        The default parameters type for the wrapper.
    available_launchers : dict
        The available launchers for the wrapper.
    """

    default_parameters = {}

    available_launchers = {
        "docker": "docker run --rm -v .:/case_dir -w /case_dir deltares/sfincs-cpu",
        "cluster": "launchSfincs.sh",
    }

    def __init__(
        self,
        templates_dir: str,
        metamodel_parameters: dict,
        fixed_parameters: dict,
        output_dir: str,
        templates_name: dict = "all",
        debug: bool = True,
    ) -> None:
        """
        Initialize the SFINCS model wrapper.
        """

        super().__init__(
            templates_dir=templates_dir,
            metamodel_parameters=metamodel_parameters,
            fixed_parameters=fixed_parameters,
            output_dir=output_dir,
            templates_name=templates_name,
            default_parameters=self.default_parameters,
        )
        self.set_logger_name(
            name=self.__class__.__name__, level="DEBUG" if debug else "INFO"
        )

    def setup_dem(self, sf: SfincsModel, case_context: dict) -> List[dict]:
        """
        Setup the DEM for the SFINCS model.
        """

        ds = sf.data_catalog.get_rasterdataset(
            data_like=case_context.get("path_to_dem_tif"),
            variables=[case_context.get("dem_tif_var_name")],
            geom=sf.region,
            meta={"version": "1"},
        )

        datasets_dep = [{"da": ds}]  # "zmin": 0.001

        sf.setup_dep(datasets_dep=datasets_dep)

        return datasets_dep

    def setup_friction(self, sf: SfincsModel, case_context: dict) -> List[dict]:
        """
        Setup the friction for the SFINCS model.
        """

        dataset_rgh = sf.data_catalog.get_rasterdataset(
            data_like=case_context.get("path_to_rgh_tif"),
        )

        datasets_rgh = [{"manning": dataset_rgh}]

        sf.setup_manning_roughness(
            datasets_rgh=datasets_rgh,
            rgh_lev_land=0,  # the minimum elevation of the land
        )

        return datasets_rgh

    def setup_infiltration(self, sf: SfincsModel, case_context: dict) -> List[dict]:
        """
        Setup the infiltration for the SFINCS model.
        """

        p_infiltration = case_context.get("path_to_inf_tif")
        ant_moisture = "avg"

        dataset_inf = sf.data_catalog.get_rasterdataset(p_infiltration)

        dataset_inf.name = "cn_{0}".format(ant_moisture)

        sf.setup_cn_infiltration(
            dataset_inf.compute(), antecedent_moisture="{0}".format(ant_moisture)
        )

        return dataset_inf

    def setup_outflow(self, sf: SfincsModel, case_context: dict) -> None:
        """
        Setup the outflow for the SFINCS model.
        """

        gdf = sf.data_catalog.get_geodataframe(
            data_like=case_context.get("path_to_outflow_shp")
        )

        sf.setup_mask_bounds(btype="outflow", include_mask=gdf, reset_bounds=True)

    def setup_waterlevel_mask(self, sf: SfincsModel, case_context: dict) -> None:
        """
        Setup the waterlevel mask for the SFINCS model.
        """

        gdf = sf.data_catalog.get_geodataframe(
            data_like=case_context.get("path_to_waterlevel_shp")
        )

        sf.setup_mask_bounds(btype="waterlevel", include_mask=gdf, reset_bounds=True)

    def set_ctimes(self, case_context: dict) -> str:
        """
        Determine the start time (TSTART) end time (TSTOP) for the simulation based on
        precipitation and water level forcing, add 1 hour,
        and return it formatted as 'YYYYMMDD HHMMSS'.
        """

        precip_forcing = case_context.get("precipitation_forcing")
        waterlevel_forcing = case_context.get("waterlevel_forcing")

        tstart_times, tstop_times = [], []

        if precip_forcing is not None:
            tstart_precip = precip_forcing["time"].values[0]
            tstop_precip = precip_forcing["time"].values[-1]

            tstart_times.append(tstart_precip)
            tstop_times.append(tstop_precip)

        if waterlevel_forcing is not None:
            tstart_waterlevel = waterlevel_forcing.index.values[0]
            tstop_waterlevel = waterlevel_forcing.index.values[-1]

            tstart_times.append(tstart_waterlevel)
            tstop_times.append(tstop_waterlevel)

        if not tstart_times:
            raise ValueError("No forcing data found to determine TSTART.")

        if not tstop_times:
            raise ValueError("No forcing data found to determine TSTOP.")

        # Get the latest time and add one hour
        tstop_max = max(tstop_times)
        tstop_plus_1h = tstop_max + np.timedelta64(1, "h")

        tstart_min = min(tstart_times)

        # Convert to pandas.Timestamp (works with numpy.datetime64) then format
        formatted_tstop = pd.to_datetime(tstop_plus_1h).strftime("%Y%m%d %H%M%S")
        formatted_tstart = pd.to_datetime(tstart_min).strftime("%Y%m%d %H%M%S")

        return formatted_tstart, formatted_tstop

    def build_template_case(self) -> None:
        """
        Build a base SFINCS model case used as a template for multiple simulations.
        This function sets up all the static components of the model that are
        common to all cases, such as grid definition, DEM, friction,
        infiltration, subgrid configuration, and boundary masks.
        """

        sf = SfincsModel(root=self.templates_dir, mode="w+")

        sf.setup_grid(
            x0=self.fixed_parameters["x0"],
            y0=self.fixed_parameters["y0"],
            dx=self.fixed_parameters["dx"],
            dy=self.fixed_parameters["dy"],
            nmax=self.fixed_parameters["nmax"],
            mmax=self.fixed_parameters["mmax"],
            rotation=self.fixed_parameters["rotation"],
            epsg=self.fixed_parameters["epsg"],
        )

        datasets_dep = self.setup_dem(sf=sf, case_context=self.fixed_parameters)

        sf.setup_mask_active(
            mask=self.fixed_parameters.get("path_to_mask"),
        )

        datasets_rgh = self.setup_friction(sf, case_context=self.fixed_parameters)

        _dataset_inf = self.setup_infiltration(sf, case_context=self.fixed_parameters)

        sf.setup_subgrid(
            datasets_dep=datasets_dep,
            datasets_rgh=datasets_rgh,
            nr_subgrid_pixels=self.fixed_parameters.get("nr_subgrid_pixels"),
            write_dep_tif=True,
            write_man_tif=False,
        )

        self.setup_outflow(sf=sf, case_context=self.fixed_parameters)

        self.setup_waterlevel_mask(sf=sf, case_context=self.fixed_parameters)

        _ = sf.plot_basemap(bmap="sat", zoomlevel=12)

        sf.write()

    def build_case(self, case_context: dict, case_dir: str) -> None:
        """
        Build the base SFINCS model. This includes setting up the grid,
        depth, friction, mask, outflow and waterlevel mask. It also
        applies the precipitation and waterlevel forcing if specified.
        """

        sf = SfincsModel(root=case_dir, mode="w+")

        sf.setup_grid(
            x0=case_context["x0"],
            y0=case_context["y0"],
            dx=case_context["dx"],
            dy=case_context["dy"],
            nmax=case_context["nmax"],
            mmax=case_context["mmax"],
            rotation=case_context["rotation"],
            epsg=case_context["epsg"],
        )
        tstart, tstop = self.set_ctimes(case_context=case_context)

        sf.config["tstop"] = tstop
        sf.config["tstart"] = tstart
        sf.config["dtout"] = 60
        sf.config["storemeteo"] = 1

        if case_context.get("quickly_waterlevel_forcing") is not None:
            """
            NOTE - There is not a specific Python function yet, 
            but one could call the setup_waterlevel_forcing function 
            twice with saving the files in between and changing their names
            """

            sf.setup_waterlevel_forcing(
                timeseries=case_context.get("quickly_waterlevel_forcing"),
                locations=case_context.get("gdf_boundary_points"),
            )
            sf.write_forcing()
            os.rename(op.join(case_dir, "sfincs.bzs"), op.join(case_dir, "sfincs.bzi"))

        if case_context.get("slowly_waterlevel_forcing") is not None:
            sf.setup_waterlevel_forcing(
                timeseries=case_context.get("slowly_waterlevel_forcing"),
                locations=case_context.get("gdf_boundary_points"),
            )
            sf.write_forcing()

        if case_context.get("precipitation_forcing") is not None:
            sf.setup_precip_forcing_from_grid(
                precip=case_context.get("precipitation_forcing"), aggregate=False
            )

        if case_context.get("gdf_crs") is not None:
            sf.setup_observation_lines(
                locations=case_context.get("gdf_crs"), merge=False
            )

        if case_context.get("gdf_obs") is not None:
            sf.setup_observation_points(locations=case_context.get("gdf_obs"))

        # if case_context.get("precipitation_forcing") is not None and case_context.get("waterlevel_forcing") is not None:
        #    self.setup_rivers(sf)
        #    sf.setup_river_inflow(
        #        rivers=case_context.get("precipitation_forcing"), keep_rivers_geom=True
        #    )

        # sf.write_forcing()

        sf.write()



from typing import List, Union

import numpy as np
from scipy import interpolate


def reef(
    dx: float,
    h0: float,
    Slope1: float,
    Slope2: float,
    Wreef: float,
    Wfore: float,
    bCrest: float,
    emsl: float,
) -> np.ndarray:
    """
    Generate a reef morphologic profile based on Pearson et al. 2017.

    This function creates a bathymetric profile for reef environments with
    distinct sections: fore reef, reef slope, reef flat, and inner slope.

    Parameters
    ----------
    dx : float
        Bathymetry mesh resolution at x axes (m)
    h0 : float
        Offshore depth (m)
    Slope1 : float
        Fore shore slope (dimensionless)
    Slope2 : float
        Inner shore slope (dimensionless)
    Wreef : float
        Reef bed width (m)
    Wfore : float
        Flume length before fore toe (m)
    bCrest : float
        Beach height (m)
    emsl : float
        Mean sea level (m)

    Returns
    -------
    np.ndarray
        Depth data values representing the reef profile (m)

    Notes
    -----
    The profile consists of several sections:
    - Fore reef: constant depth at offshore level
    - Reef slope: linear slope from offshore to reef flat
    - Reef flat: constant depth at mean sea level
    - Inner slope: linear slope from reef flat to beach
    - Plane beach: gentle slope for overtopping dissipation
    """

    # flume length
    W_inner = bCrest / Slope2
    W1 = int(abs(h0 - emsl) / Slope1)

    # sections length
    x1 = np.arange(0, Wfore, dx)
    x2 = np.arange(0, W1, dx)
    x3 = np.arange(0, Wreef, dx)
    x4 = np.arange(0, W_inner, dx)

    # curve equation
    y_fore = np.zeros(len(x1)) + [h0]
    y1 = -Slope1 * x2 + h0
    y2 = np.zeros(len(x3)) + emsl
    y_inner = -Slope2 * x4 + emsl

    # overtopping cases: an inshore plane beach to dissipate overtopped flux
    plane = 0.005 * np.arange(0, 150, 1) + y_inner[-1]

    # concatenate depth
    depth = np.concatenate([y_fore, y1, y2, y_inner, plane])

    return depth


def linear(dx: float, h0: float, bCrest: float, m: float, Wfore: float) -> np.ndarray:
    """
    Generate a simple linear profile (y = m * x + n).

    This function creates a bathymetric profile with a constant slope
    from offshore to the beach crest.

    Parameters
    ----------
    dx : float
        Bathymetry mesh resolution at x axes (m)
    h0 : float
        Offshore depth (m)
    bCrest : float
        Beach height (m)
    m : float
        Profile slope (dimensionless)
    Wfore : float
        Flume length before slope toe (m)

    Returns
    -------
    np.ndarray
        Depth data values representing the linear profile (m)

    Notes
    -----
    The profile consists of:
    - Fore section: constant depth at offshore level
    - Main slope: linear slope from offshore to beach
    - Beach slope: linear slope from sea level to beach crest
    - Plane beach: gentle slope for overtopping dissipation
    """

    # Flume length
    W1 = int(h0 / m)
    W2 = int(bCrest / m)

    # Sections length
    x1 = np.arange(0, Wfore, dx)
    x2 = np.arange(0, W1, dx)
    x3 = np.arange(0, W2, dx)

    # Curve equation
    y_fore = np.zeros(len(x1)) + [h0]
    y1 = -m * x2 + h0
    y2 = -m * x3

    # Overtopping cases: an inshore plane beach to dissipate overtopped flux
    plane = 0.005 * np.arange(0, len(y2), 1) + y2[-1]  # Length bed = 2 L

    # concatenate depth
    depth = np.concatenate([y_fore, y1, y2, plane])

    return depth


def parabolic(
    dx: float, h0: float, A: float, xBeach: float, bCrest: float
) -> np.ndarray:
    """
    Generate a parabolic profile (y = A * x^(2/3)).

    This function creates a bathymetric profile following the equilibrium
    beach profile theory with a parabolic shape.

    Parameters
    ----------
    dx : float
        Bathymetry mesh resolution at x axes (m)
    h0 : float
        Offshore depth (m)
    A : float
        Parabola coefficient (m^(1/3))
    xBeach : float
        Beach length (m)
    bCrest : float
        Beach height (m)

    Returns
    -------
    np.ndarray
        Depth data values representing the parabolic profile (m)

    Notes
    -----
    The profile follows the equilibrium beach profile theory where
    depth varies as x^(2/3) from the shoreline to the closure depth.
    The beach section is linear from the shoreline to the beach crest.
    """

    lx = np.arange(1, xBeach, dx)
    y = -(bCrest / xBeach) * lx

    depth, xl = [], []
    x, z = 0, 0

    while z <= h0:
        z = A * x ** (2 / 3)
        depth.append(z)
        xl.append(x)
        x += dx

    f = interpolate.interp1d(xl, depth)
    xnew = np.arange(0, int(np.round(len(depth) * dx)), 1)
    ynew = f(xnew)

    # concatenate depth
    depth = np.concatenate([ynew[::-1], y])

    return depth


def biparabolic(
    h0: float, hsig: float, omega_surf_list: Union[float, np.ndarray], TR: float
) -> np.ndarray:
    """
    Generate a biparabolic profile based on Bernabeu et al. 2013.

    This function creates a bathymetric profile with two parabolic sections
    separated by a discontinuity point, suitable for mixed-sediment beaches.

    Parameters
    ----------
    h0 : float
        Offshore water level (m)
    hsig : float
        Significant wave height (m)
    omega_surf_list : float or np.ndarray
        Intertidal dimensionless fall velocity (1 <= omega_surf <= 5)
    TR : float
        Tidal range (m)

    Returns
    -------
    np.ndarray
        Depth data values representing the biparabolic profile (m)

    Notes
    -----
    The biparabolic profile consists of:
    - Lower section: parabolic profile for fine sediments
    - Upper section: parabolic profile for coarse sediments
    - Discontinuity point: transition between the two sections
    - The profile is centered on mean tide level

    The empirical parameters A, B, C, D are adjusted based on the
    dimensionless fall velocity parameter.
    """

    # Discontinuity point
    hr = 1.1 * hsig + TR

    # Legal point
    _ha = 3 * hsig + TR

    # Empirical adjusted parameters
    A = 0.21 - 0.02 * omega_surf_list
    B = 0.89 * np.exp(-1.24 * omega_surf_list)
    C = 0.06 + 0.04 * omega_surf_list
    D = 0.22 * np.exp(-0.83 * omega_surf_list)

    # Different values for the height
    h = np.linspace(0, h0, 150)
    h_cont = []

    # Important points for the profile
    _xr = (hr / A) ** (3 / 2) + (B / (A ** (3 / 2))) * hr**3

    # Lines for the profile
    x, Xx, X, xX = [], [], [], []

    for hs in h:  # For each vertical point
        if hs < hr:
            x_max = 0
            xapp = (hs / A) ** (3 / 2) + (B / (A ** (3 / 2))) * hs**3
            x.append(xapp)
            x_max = max(xapp, x_max)
            if hs > (hr - 1.5):
                Xxapp = (hs / C) ** (3 / 2) + (D / (C ** (3 / 2))) * hs**3
                Xx.append(Xxapp)
                h_cont.append(hs)
        else:
            Xapp = (hs / C) ** (3 / 2) + (D / (C ** (3 / 2))) * hs**3
            if (hs - hr) < 0.1:
                x_diff = x_max - Xapp
            X.append(Xapp)
            if hs < (hr + 1.5):
                xXapp = (hs / A) ** (3 / 2) + (B / (A ** (3 / 2))) * hs**3
                xX.append(xXapp)
                h_cont.append(hs)

    h_cont = np.array(h_cont)
    x_tot = np.concatenate((np.array(x), np.array(X) + x_diff))
    # x_cont = np.concatenate((np.array(Xx)+x_diff, np.array(xX)))

    # Centering the y-axis in the mean tide
    xnew = np.arange(0, x_tot[-1], 1)
    # xnew_border = np.arange(x_tot[-1]-x_cont[0], x_cont[-1]-x_cont[-1], 1)
    depth = h - TR / 2
    # border = (-h_cont+TR/2)

    f = interpolate.interp1d(x_tot, depth)
    # f1 = interpolate.interp1d(x_cont, border)
    ynew = f(xnew)[::-1]
    # ynew_border = f1(xnew_border)[::-1]

    depth = (h - TR / 2)[::-1]
    # border = (-h_cont+TR/2)[::-1]

    # plot
    # TODO: move plot to plots.py
    # fig, ax = plt.subplots(1, figsize = (12, 4))
    # ax.plot(xnew, -ynew, color='k', zorder=3)
    # ax.fill_between(xnew, np.zeros((len(xnew)))+(-ynew[0]),
    #                -ynew,facecolor="wheat", alpha=1, zorder=2)
    # ax.scatter(x_tot[-1]-xr, -hr+TR/2, s=30, c='red', label='Discontinuity point', zorder=5)
    # ax.fill_between(xnew, -ynew, np.zeros(len(xnew)), facecolor="deepskyblue", alpha=0.5, zorder=1)
    # ax.axhline(-ha+TR/2, color='grey', ls='-.', label='Available region')
    # ax.axhline(TR/2, color='silver', ls='--', label='HT')
    # ax.axhline(0, color='lightgrey', ls='--', label='MSL')
    # ax.axhline(-TR/2, color='silver', ls='--', label='LT')
    # ax.scatter(xnew_border, -ynew_border, c='k', s=1, marker='_', zorder=4)

    # attrbs
    # ax.set_ylim(-ynew[0], -ynew[-1]+1)
    # ax.set_xlim(0, x_tot[-1])
    # set_title  = '$\Omega_{sf}$ = ' + str(omega_surf_list)
    # set_title += ', TR = ' + str(TR)
    # ax.set_title(set_title)
    # ax.legend(loc='upper left')
    # ax.set_ylabel('$Depth$ $[m]$', fontweight='bold')
    # ax.set_xlabel('$X$ $[m]$', fontweight='bold')

    # TODO: deph or ynew ?
    return ynew


def custom_profile(
    dx: float,
    emsl: float,
    xs: Union[List[float], np.ndarray],
    ys: Union[List[float], np.ndarray],
) -> np.ndarray:
    """
    Generate a custom N-point profile from user-defined coordinates.

    This function creates a bathymetric profile by interpolating between
    user-specified x,y coordinate pairs.

    Parameters
    ----------
    dx : float
        Bathymetry mesh resolution at x axes (m)
    emsl : float
        Mean sea level (m) - used for reference but not directly applied
    xs : list or np.ndarray
        X coordinate values (m)
    ys : list or np.ndarray
        Y coordinate values (m) - positive values represent elevation above MSL

    Returns
    -------
    np.ndarray
        Depth data values representing the custom profile (m)

    Notes
    -----
    The function uses linear interpolation between the provided points.
    The output depths are negative values (below sea level) as is
    conventional for bathymetric data.

    The xs and ys arrays must have the same length and be sorted
    in increasing x order for proper interpolation.
    """

    # flume length
    xnew = np.arange(xs[0], xs[-1], dx)
    f = interpolate.interp1d(xs, ys)
    ynew = f(xnew)

    depth = -ynew

    return depth



import re
from collections import defaultdict
from copy import deepcopy
from typing import Dict, List, Tuple

import cartopy.crs as ccrs
import geopandas as gpd
import matplotlib.pyplot as plt
import numpy as np

# import ocsmesh
import pandas as pd
import rasterio

# from jigsawpy.msh_t import jigsaw_msh_t
from matplotlib.axes import Axes
from pyproj.enums import TransformDirection
from rasterio.mask import mask
from shapely.geometry import LineString, MultiLineString, Polygon, mapping
from shapely.ops import transform

from ..core.geo import buffer_area_for_polygon
from ..core.plotting.colors import hex_colors_land, hex_colors_water
from ..core.plotting.utils import join_colormaps


def plot_mesh_edge(
    msh_t: jigsaw_msh_t, ax: Axes = None, to_geo: callable = None, **kwargs
) -> None:
    """
    Plots the edges of a triangular mesh on a given set of axes.

    Parameters
    ----------
    msh_t : jigsaw_msh_t
        An object containing mesh data. It must have:
        - 'vert2['coord']' containing the coordinates of the mesh vertices
        - 'tria3['index']' containing the indices of the triangles
    ax : Axes, optional
        The axes to plot on. If None, a new plot is created. Default is None.
    to_geo : callable, optional
        A function to transform (x, y) coordinates from projected to geographic CRS.
    **kwargs : keyword arguments, optional
        Additional keyword arguments passed to the `triplot` function.
        These can be used to customize the plot (e.g., color, line style).
    """

    crd = np.array(msh_t.vert2["coord"], copy=True)
    cnn = msh_t.tria3["index"]

    if to_geo is not None:
        crd[:, 0], crd[:, 1] = to_geo(crd[:, 0], crd[:, 1])
        bnd = [crd[:, 0].min(), crd[:, 0].max(), crd[:, 1].min(), crd[:, 1].max()]

    ax.triplot(crd[:, 0], crd[:, 1], cnn, **kwargs)
    ax.set_extent([*bnd], crs=ccrs.PlateCarree())
    gl = ax.gridlines(draw_labels=True)
    gl.top_labels = False
    gl.right_labels = False


def plot_mesh_vals(
    msh_t: jigsaw_msh_t,
    ax: Axes = None,
    colorbar: bool = True,
    clim: Tuple[float, float] = None,
    to_geo: callable = None,
    **kwargs,
) -> Axes:
    """
    Plots the mesh values on a triangular mesh, with optional transformation
    from UTM to geographic coordinates.

    Parameters
    ----------
    msh_t : jigsaw_msh_t
        An object containing the mesh data. Must include:
        - vert2['coord']: coordinates of mesh vertices (N, 2)
        - tria3['index']: triangle connectivity (M, 3)
        - value: values to plot (length M or Mx1)
    ax : matplotlib Axes, optional
        Axes to draw on. If None, a new one is created.
    colorbar : bool, optional
        If True, show colorbar. Default is True.
    clim : tuple, optional
        Color limits (vmin, vmax). If None, autoscale.
    to_geo : callable, optional
        Function to transform (x, y) in projected coordinates to (lon, lat),
    **kwargs : additional keyword args for tricontourf

    Returns
    -------
    ax : matplotlib Axes
        The axes with the plot.
    """

    crd = np.array(msh_t.vert2["coord"], copy=True)
    cnn = msh_t.tria3["index"]
    val = msh_t.value.flatten()

    if to_geo is not None:
        crd[:, 0], crd[:, 1] = to_geo(crd[:, 0], crd[:, 1])

    if ax is None:
        _, ax = plt.subplots()

    mappable = ax.tricontourf(crd[:, 0], crd[:, 1], cnn, val, **kwargs)

    if colorbar:
        if clim is not None:
            mappable.set_clim(*clim)
        cbar = plt.colorbar(mappable, ax=ax)
        cbar.set_label("Mesh spacing conditioning (m)")

    gl = ax.gridlines(draw_labels=True)
    gl.top_labels = False
    gl.right_labels = False
    return ax


def plot_bathymetry(rasters_path: List[str], polygon: Polygon, ax: Axes) -> Axes:
    """
    Plots bathymetric raster data and overlays a polygon on top of it.

    Parameters
    ----------
    rasters_path : List[str]
        A list of file paths to the raster files.
    polygon : Polygon
        A polygon to overlay on the raster data.
    ax : Axes
        The axes on which to plot the data.

    Returns
    -------
    ax : Axes
        The axes object with the plotted raster data and polygon.
    """

    data = []
    for path in rasters_path:
        with rasterio.open(path) as src:
            raster_data = src.read(1)
            no_data_value = src.nodata
            if no_data_value is not None:
                raster_data = np.ma.masked_equal(raster_data, no_data_value)
            data.append(raster_data)
            transform = src.transform

    x_polygon, y_polygon = polygon.exterior.xy

    height, width = data[0].shape
    cols, rows = np.meshgrid(np.arange(width), np.arange(height))
    xs, ys = rasterio.transform.xy(transform, rows, cols)

    vmin = np.nanmin(data[0])
    vmax = np.nanmax(data[0])

    cmap, norm = join_colormaps(
        cmap1=hex_colors_water,
        cmap2=hex_colors_land,
        value_range1=(vmin, 0.0),
        value_range2=(0.0, vmax),
        name="raster_cmap",
    )

    im = ax.imshow(
        data[0],
        cmap=cmap,
        norm=norm,
        extent=(np.min(xs), np.max(xs), np.min(ys), np.max(ys)),
    )
    cbar = plt.colorbar(im, ax=ax)
    cbar.set_label("Depth (m)")
    ax.set_title("Raster")

    ax.plot(x_polygon, y_polygon, color="red", linewidth=1)
    gl = ax.gridlines(draw_labels=True)
    gl.top_labels = False
    gl.right_labels = False

    return ax


def clip_bathymetry(
    input_raster_paths: List[str],
    output_path: str,
    domain: Polygon,
    margin: float,
) -> float:
    """
    Clips bathymetric raster data using a specified domain polygon,
    applies a margin buffer, and saves the clipped raster.

    Parameters
    ----------
    input_raster_paths : List[str]
        List of file paths to the input raster files.
    output_path : str
        Destination file path to save the clipped raster.
    domain : Polygon
        Polygon geometry used to clip the raster.
    margin : float
        Buffer factor applied to the domain before clipping.

    Returns
    -------
    float
        The mean resolution of the raster.
    """

    buffered_polygon = buffer_area_for_polygon(domain, margin)

    for path in input_raster_paths:
        with rasterio.open(path) as src:
            domain_gdf = gpd.GeoDataFrame(
                index=[0], geometry=[buffered_polygon], crs="EPSG:4326"
            ).to_crs(src.crs)

            out_image, out_transform = mask(
                src, [mapping(domain_gdf.geometry[0])], crop=True
            )

            out_meta = src.meta.copy()
            out_meta.update(
                {
                    "driver": "GTiff",
                    "height": out_image.shape[1],
                    "width": out_image.shape[2],
                    "transform": out_transform,
                }
            )
        mean_raster_resolution = get_raster_resolution(path)
    with rasterio.open(output_path, "w", **out_meta) as dest:
        dest.write(out_image)
    return mean_raster_resolution


def get_raster_resolution(raster_path: str) -> float:
    """
    Get the mean resolution of a raster in meters.

    Parameters
    ----------
    raster_path : str
        Path to the raster file.
    Returns
    -------
    float
        Mean resolution of the raster in meters.
    ----------
    Notes
    This function uses rasterio to open the raster file and extract its resolution.
    The mean resolution is calculated as the average of the absolute values of the x and y resolutions.
    """

    with rasterio.open(raster_path) as src:
        res_x, res_y = src.res
        mean_resolution = (abs(res_x) + abs(res_y)) / 2
    return mean_resolution


def clip_bati_manning(
    rasters_path: List[str],
    output_path: str,
    domain: Polygon,
    mas: float,
    UTM: bool,
    manning: float,
) -> None:
    """
    Clips bathymetric raster data using a specified domain polygon and applies
    Manning's coefficient.

    Parameters
    ----------
    rasters_path : List[str]
        A list of file paths to the raster files to be clipped.
    output_path : str
        The file path to save the clipped raster data.
    domain : Polygon
        The domain polygon used to clip the rasters.
    mas : float
        A buffer factor applied to the domain polygon based on its area and length.
    UTM : bool
        If True, assumes the coordinate reference system is EPSG:4326;
        otherwise, assumes EPSG:32630 (UTM projection).
    manning : float
        The Manning's coefficient to apply to the raster data.
    """

    original_polygon = buffer_area_for_polygon(domain, mas)

    if UTM:
        crrs = "EPSG:4326"
    else:
        crrs = "EPSG:32630"
    for path in rasters_path:
        with rasterio.open(path) as src:
            gdf_polygon = gpd.GeoDataFrame(
                index=[0], geometry=[original_polygon], crs=crrs
            )
            gdf_polygon = gdf_polygon.to_crs(src.crs)

            out_image, out_transform = mask(
                src, [mapping(gdf_polygon.geometry[0])], crop=True
            )

            out_meta = src.meta.copy()
            out_meta.update(
                {
                    "driver": "GTiff",
                    "height": out_image.shape[1],
                    "width": out_image.shape[2],
                    "transform": out_transform,
                }
            )

    out_image = np.ones(out_image.shape) * (-manning)
    with rasterio.open(output_path, "w", **out_meta) as dest:
        dest.write(out_image)


def plot_boundaries(mesh: jigsaw_msh_t, ax: Axes, to_geo: callable = None) -> None:
    """
    Plots the boundaries of a mesh, including ocean, interior (islands), and land areas.

    Parameters
    ----------
    mesh : jigsaw_msh_t
        The mesh object containing the mesh data and boundaries.
    ax : Axes
        The axes on which to plot the boundaries.
    to_geo : callable, optional
        A function to transform coordinates from projected to geographic CRS.
    """

    plot_mesh_edge(mesh.msh_t, to_geo=to_geo, ax=ax, color="gray", lw=0.5)

    def plot_boundary(gdf, color, label):
        try:
            if to_geo:
                gdf = gdf.copy()
                gdf["geometry"] = gdf["geometry"].apply(
                    lambda geom: transform(to_geo, geom)
                )
            gdf.plot(ax=ax, color=color, label=label)
        except Exception as _e:
            print(f"No {label} boundaries available")

    plot_boundary(mesh.boundaries.ocean(), color="b", label="Ocean")
    plot_boundary(mesh.boundaries.interior(), color="g", label="Islands")
    plot_boundary(mesh.boundaries.land(), color="r", label="Land")

    ax.set_title("Mesh Boundaries")
    ax.legend()


def plot_bathymetry_interp(mesh: jigsaw_msh_t, to_geo, ax: Axes) -> None:
    """
    Plots the interpolated bathymetry data on a mesh.

    Parameters
    ----------
    mesh : jigsaw_msh_t
        The mesh object containing the bathymetry values and mesh structure.
    ax : Axes
        The axes on which to plot the interpolated bathymetry.
    to_geo : callable
        A function to transform coordinates from projected to geographic CRS.
    """

    crd = np.array(mesh.msh_t.vert2["coord"], copy=True)

    if to_geo is not None:
        crd[:, 0], crd[:, 1] = to_geo(crd[:, 0], crd[:, 1])
        bnd = [crd[:, 0].min(), crd[:, 0].max(), crd[:, 1].min(), crd[:, 1].max()]

    triangle = mesh.msh_t.tria3["index"]
    Z = np.mean(mesh.msh_t.value.flatten()[triangle], axis=1)
    vmin = np.nanmin(Z)
    vmax = np.nanmax(Z)

    cmap, norm = join_colormaps(
        cmap1=hex_colors_water,
        cmap2=hex_colors_land,
        value_range1=(vmin, 0.0),
        value_range2=(0.0, vmax),
        name="raster_cmap",
    )

    im = ax.tripcolor(
        crd[:, 0],
        crd[:, 1],
        triangle,
        facecolors=Z,
        cmap=cmap,
        norm=norm,
    )
    ax.set_title("Interpolated Bathymetry")
    gl = ax.gridlines(draw_labels=True)
    gl.top_labels = False
    gl.right_labels = False
    cbar = plt.colorbar(im, ax=ax)
    cbar.set_label("Depth (m)")
    ax.set_extent([*bnd], crs=ccrs.PlateCarree())


def simply_polygon(base_shape: Polygon, simpl_UTM: float, project) -> Polygon:
    """
    Simplifies a polygon by transforming it to a projected coordinate system (e.g., UTM),
    applying geometric simplification, and transforming it back to geographic coordinates.

    Parameters
    ----------
    base_shape : Polygon
        The input polygon in geographic coordinates (e.g., WGS84).
    simpl_UTM : float
        Tolerance for simplification (in projected units, typically meters). Higher values result in simpler shapes.
    project : pyproj.Transformer.transform
        A projection function that converts coordinates from geographic to projected (e.g., WGS84 → UTM).

    Returns
    -------
    Polygon
        The simplified polygon in geographic coordinates.
    """

    base_shape_utm = transform(project, base_shape)

    simple_shape_utm = base_shape_utm.simplify(simpl_UTM)

    simple_shape = transform(
        lambda x, y: project(x, y, direction=TransformDirection.INVERSE),
        simple_shape_utm,
    )

    return simple_shape


def remove_islands(base_shape: Polygon, threshold_area: float, project) -> Polygon:
    """
    Transforms a polygon to a projected coordinate system (e.g., UTM), filters out small interior rings
    (holes) based on a minimum area threshold, and transforms the simplified polygon back to geographic coordinates.

    Parameters
    ----------
    base_shape : Polygon
        The input polygon in geographic coordinates (e.g., WGS84).
    threshold_area : float
        Minimum area (in projected units, e.g., square meters) for interior rings (holes) to be preserved.
        Interior rings smaller than this threshold will be removed.
    project : callable
        A projection function, typically `pyproj.Transformer.transform`, that converts coordinates
        from geographic to projected CRS.

    Returns
    -------
    Polygon
        The polygon with small interior rings removed, transformed back to geographic coordinates.
    """

    base_shape_utm = transform(project, base_shape)
    interior_shape_utm = [
        interior
        for interior in base_shape_utm.interiors
        if Polygon(interior).area >= threshold_area
    ]
    simple_shape_utm = Polygon(base_shape_utm.exterior, interior_shape_utm)
    simple_shape = transform(
        lambda x, y: project(x, y, direction=TransformDirection.INVERSE),
        simple_shape_utm,
    )

    return simple_shape


def decode_open_boundary_data(data: List[str]) -> dict:
    """
    Decodes open boundary data from a given list of strings and
    returns a dictionary containing boundary information.

    Parameters
    ----------
    data : List[str]
        List of strings containing boundary data.

    Returns
    -------
    dict
        A dictionary with keys corresponding to open boundary identifiers (e.g., 'open_boundary_1')
        and values as lists of integers representing boundary node indices.

    Examples
    --------
    >>> data = [
    ...     "100! 200! 300!",
    ...     "open_boundary_1",
    ...     "open_boundary_2",
    ...     "open_boundary_3",
    ...     "land boundaries",
    ...     "open_boundary_1! 10",
    ...     "open_boundary_2! 20",
    ...     "open_boundary_3! 30",
    ... ]
    >>> boundaries = decode_open_boundary_data(data)
    >>> print(boundaries)
    {'open_boundary_1': [10], 'open_boundary_2': [20], 'open_boundary_3': [30]}
    """

    N_obd = int(data[0].split("!")[0])
    boundary_info = {}
    key = data[2][-16:-1]
    boundary_info[key] = []

    for line in data[3:-1]:
        line = line.strip()
        if "!" not in line:
            N = int(line)
            boundary_info[key].append(N)
        else:
            if "land boundaries" in line:
                if len(boundary_info) != N_obd:
                    print("reading error")
                return boundary_info
            match = re.search(r"open_boundary_\d+", line)
            key = match.group(0)
            boundary_info[key] = []

    return boundary_info


def compute_circumcenter(p0: np.ndarray, p1: np.ndarray, p2: np.ndarray) -> np.ndarray:
    """
    Technical Reference Manual D-Flow Flexible Mesh 13 May 2025 Revision: 80268
    Compute the circumcenter of a triangle from its 3 vertices.
    Ref: Figure 3.12, Equation (3.6), D-Flow Flexible Mesh Technical Reference Manual.

    Parameters
    ----------
    p0, p1, p2 : np.ndarray
        2D coordinates of the triangle vertices.

    Returns
    -------
    np.ndarray
        2D coordinates of the circumcenter.
    """

    A = p1 - p0
    B = p2 - p0
    AB_perp = np.array([A[1], -A[0]])
    AC_perp = np.array([B[1], -B[0]])
    mid_AB = (p0 + p1) / 2
    mid_AC = (p0 + p2) / 2
    M = np.array([AB_perp, -AC_perp]).T
    b = mid_AC - mid_AB
    try:
        t = np.linalg.solve(M, b)
        center = mid_AB + t[0] * AB_perp
    except np.linalg.LinAlgError:
        center = (p0 + p1 + p2) / 3
    return center


def build_edge_to_cells(elements: np.ndarray) -> Dict[Tuple[int, int], List[int]]:
    """
    Technical Reference Manual D-Flow Flexible Mesh 13 May 2025 Revision: 80268
    Build edge -> list of adjacent element indices (max 2).
    Ref: Connectivity structure implied in Section 3.5.1 and Fig 3.2a.

    Parameters
    ----------
    elements : np.ndarray
        Array of triangle elements (indices of vertices).

    Returns
    -------
    edge_to_cells : Dict[Tuple[int, int], List[int]]
        Dictionary mapping edges to the list of adjacent element indices.
    """

    edge_to_cells = defaultdict(list)
    for idx, elem in enumerate(elements):
        for i in range(3):
            a = elem[i]
            b = elem[(i + 1) % 3]
            edge = tuple(sorted((a, b)))
            edge_to_cells[edge].append(idx)

    return edge_to_cells


def detect_circumcenter_too_close(
    X: np.ndarray, Y: np.ndarray, elements: np.ndarray, aj_threshold: float = 0.1
) -> np.ndarray:
    """
    Technical Reference Manual D-Flow Flexible Mesh 13 May 2025 Revision: 80268
    Detect elements where the distance between adjacent circumcenters is small compared
    to the shared edge length: aj = ||xRj - xLj|| / ||x1j - x0||
    Ref: Equation (3.6) in Section 3.5.1 (Grid Orthogonalization), D-Flow Flexible Mesh Manual.

    Parameters
    ----------
    X, Y : np.ndarray
        1D arrays of x and y coordinates of the nodes.
    elements : np.ndarray
        2D array of shape (nelmts, 3) containing the node indices for each triangle element.
    aj_threshold : float, optional
        Threshold for the ratio of circumcenter distance to edge length. Default is 0.1.

    Returns
    -------
    bad_elements_mask : np.ndarray
        Boolean mask indicating which elements are problematic (True if bad).
    """

    nodes = np.column_stack((X, Y))
    centers = np.array(
        [
            compute_circumcenter(nodes[i0], nodes[i1], nodes[i2])
            for i0, i1, i2 in elements
        ]
    )  # Ref: Fig 3.12

    edge_to_cells = build_edge_to_cells(elements)
    bad_elements_mask = np.zeros(len(elements), dtype=bool)

    for edge, cells in edge_to_cells.items():
        if len(cells) != 2:
            continue  # Internal edges only (Ref: Ji in Eq. 3.7)

        idx0, idx1 = cells
        c0 = centers[idx0]  # x_Lj
        c1 = centers[idx1]  # x_Rj
        node0 = nodes[edge[0]]  # x0
        node1 = nodes[edge[1]]  # x1j

        edge_length = np.linalg.norm(node1 - node0)  # Denominator of aj (||x1j - x0||)
        center_dist = np.linalg.norm(c1 - c0)  # Numerator of aj (||xRj - xLj||)

        aj = center_dist / edge_length if edge_length > 0 else 0  # Equation (3.6)

        if aj < aj_threshold:
            # If the ratio is too low, mark both elements as problematic
            bad_elements_mask[idx0] = True
            bad_elements_mask[idx1] = True

    return bad_elements_mask


def define_mesh_target_size(
    rasters: List[rasterio.io.DatasetReader],
    raster_resolution_meters: float,
    depth_ranges: dict,
    nprocs: int = 1,
) -> ocsmesh.Hfun:
    """
    Define the mesh target size based on depth ranges and their corresponding values.

    Parameters
    ----------
    rasters : List[rasterio.io.DatasetReader]
        List of raster objects.
    raster_resolution_meters : float
        Resolution of the raster in meters.
    depth_ranges : dict, optional
        Dictionary containing depth ranges and their corresponding mesh sizes and rates.
        Format: {(lower_bound, upper_bound): {'value': mesh_size, 'rate': expansion_rate}}
    nprocs : int, optional
        Number of processors to use for the mesh generation. Default is 1.

    Returns
    -------
    mesh_spacing : ocsmesh.Hfun
        Hfun object with the defined mesh target size.
    """

    if depth_ranges is None:
        # Default depth-to-mesh size mapping
        depth_ranges = {
            (-200_000, -250): {"value": 26_000, "rate": 0.0001},  # Very deep ocean
            (-250, -100): {"value": 13_000, "rate": 0.0001},  # Continental slope
            (-100, -75): {"value": 6_500, "rate": 0.0001},  # Outer shelf
            (-75, -25): {"value": 3_250, "rate": 0.0001},  # Mid shelf
            (-25, 2.5): {"value": 1_700, "rate": 0.0001},  # Coastal zone
        }

    all_values = [zone["value"] for zone in depth_ranges.values()]

    points_by_cell = 1  # Number of depth points per minimum size cell for the final cell size definition

    rasters_copy = deepcopy(rasters)
    for raster in rasters_copy:
        raster.resample(
            scaling_factor=raster_resolution_meters * points_by_cell / min(all_values)
        )

    mesh_spacing = ocsmesh.Hfun(
        rasters_copy,
        hmin=min(all_values),
        hmax=max(all_values),
        nprocs=nprocs,
    )

    for (lower, upper), params in depth_ranges.items():
        mesh_spacing.add_topo_bound_constraint(
            value=params["value"],
            lower_bound=lower,
            upper_bound=upper,
            value_type="max",
            rate=params["rate"],
        )

    return mesh_spacing


def read_lines(poly_line: str) -> MultiLineString:
    """
    Reads a CSV file containing coordinates of a polyline and returns a MultiLineString.
    The CSV file should have two columns for x and y coordinates, with NaN values indicating breaks in the line.
    Parameters
    ----------
    poly_line : str
        Path to the CSV file containing the polyline coordinates
    Returns
    -------
    MultiLineString
        A MultiLineString object representing the polyline segments
    """

    coords_line = pd.read_csv(poly_line, sep=",", header=None)
    segments = []
    current_segment = []
    for index, row in coords_line.iterrows():
        if row.isna().any():
            if current_segment:
                segments.append(LineString(current_segment))
                current_segment = []
        else:
            current_segment.append(tuple(row))

    if current_segment:
        segments.append(LineString(current_segment))
    return MultiLineString(segments)


def get_raster_resolution_meters(lon_center, lat_center, raster_resolution, project):
    """
    Calculate the raster resolution in meters based on the center coordinates and the raster resolution in degrees.

    Parameters
    ----------
    lon_center : float
        Longitude of the center point.
    lat_center : float
        Latitude of the center point.
    raster_resolution : float
        Raster resolution in degrees.
    Returns
    -------
    float
        Raster resolution in meters.
    """
    x_center, y_center = project(lon_center, lat_center)
    x_center_raster_resolution, y_center_raster_resolution = project(
        lon_center + raster_resolution / np.sqrt(2),
        lat_center + raster_resolution / np.sqrt(2),
    )
    raster_resolution_meters = np.mean(
        [
            abs(x_center - x_center_raster_resolution),
            abs(y_center - y_center_raster_resolution),
        ]
    )
    return raster_resolution_meters



from typing import Any, Dict

import cmocean
import matplotlib.colors as mcolors
import numpy as np
from matplotlib import cm
from matplotlib import pyplot as plt
from matplotlib.colors import ListedColormap

default_colors = [
    "#636EFA",
    "#EF553B",
    "#00CC96",
    "#AB63FA",
    "#FFA15A",
    "#19D3F3",
    "#FF6692",
    "#B6E880",
    "#FF97FF",
    "#FECB52",
]

hex_colors_water = [
    "#4a84b5",
    "#5493c8",
    "#5fa9d1",
    "#74c3dc",
    "#8ed7e8",
    "#a0e2ef",
    "#b7f1eb",
    "#c8ebd8",
    "#d7e8c3",
    "#e2e5a5",
    "#f4cda0",
    "#f1e2c6",
]

hex_colors_land = [
    "#cfe2bd",
    "#aece91",
    "#7eb14f",
    "#76ac44",
    "#6ea739",
    "#66a22e",
    "#518134",
    "#518134",
    "#3E731E",
]


scatter_defaults = {
    "figsize": (9, 8),
    "marker": ".",
    "color_data": "#00CC96",
    "color_subset": "#AB63FA",
    "alpha_data": 0.5,
    "alpha_subset": 0.7,
    "size_data": 10,
    "size_centroid": 70,
    "fontsize": 12,
}


def get_config_variables() -> Dict[str, Dict[str, Any]]:
    """
    Get configuration variables for different meteorological parameters.

    Returns
    -------
    Dict[str, Dict[str, Any]]
        Dictionary containing configuration for different variables:
        - geo500hpa: Geopotential height at 500hPa
        - mslp: Mean Sea Level Pressure
        - mslp_grad: MSLP gradient
        - sst: Sea Surface Temperature
        - other: Default configuration
    """

    config_variables = {}

    config_variables["geo500hpa"] = {
        "cmap": "Blues",
        "limits": (45000, 60000),
        "label": "Geo500Hpa [m]",
    }

    config_variables["mslp"] = {
        "cmap": "RdBu_r",
        "limits": (1014 - 20, 1014 + 20),
        "label": "MSLP [mbar]",
    }

    config_variables["mslp_grad"] = {
        "cmap": "BuGn",
        "limits": (0, 100),
        "label": "MSLP - grad [mbar]",
    }

    config_variables["sst"] = {
        "cmap": "hot_r",
        "limits": (0, 30),
        "label": "SST - [°C]",
    }

    config_variables["other"] = {
        "cmap": "rainbow",
        "limits": (None, None),
        "label": " ",
    }

    return config_variables


def colors_awt() -> np.ndarray:
    """
    Get colors for Annual Weather Types (6 categories).

    Returns
    -------
    np.ndarray
        Array of RGB colors for 6 AWT categories
    """

    l_colors_dwt = [
        (155 / 255.0, 0, 0),
        (1, 0, 0),
        (255 / 255.0, 216 / 255.0, 181 / 255.0),
        (164 / 255.0, 226 / 255.0, 231 / 255.0),
        (0 / 255.0, 190 / 255.0, 255 / 255.0),
        (51 / 255.0, 0 / 255.0, 207 / 255.0),
    ]

    return np.array(l_colors_dwt)


def colors_mjo() -> np.ndarray:
    """
    Get colors for MJO 25 categories.

    Returns
    -------
    np.ndarray
        Array of RGB colors for 25 MJO categories.
    """

    l_named_colors = [
        "lightskyblue",
        "deepskyblue",
        "royalblue",
        "mediumblue",
        "darkblue",
        "darkblue",
        "darkturquoise",
        "turquoise",
        "maroon",
        "saddlebrown",
        "chocolate",
        "gold",
        "orange",
        "orangered",
        "red",
        "firebrick",
        "Purple",
        "darkorchid",
        "mediumorchid",
        "magenta",
        "mediumslateblue",
        "blueviolet",
        "darkslateblue",
        "indigo",
        "darkgray",
    ]

    np_colors_rgb = np.array([mcolors.to_rgb(c) for c in l_named_colors])

    return np_colors_rgb


def colors_dwt(num_clusters: int) -> np.ndarray:
    """
    Get colors for Daily Weather Types.

    Parameters
    ----------
    num_clusters : int
        Number of clusters to get colors for.

    Returns
    -------
    np.ndarray
        Array of RGB colors for the specified number of clusters.
    """

    l_colors_dwt = [
        (1.0000, 0.1344, 0.0021),
        (1.0000, 0.2669, 0.0022),
        (1.0000, 0.5317, 0.0024),
        (1.0000, 0.6641, 0.0025),
        (1.0000, 0.9287, 0.0028),
        (0.9430, 1.0000, 0.0029),
        (0.6785, 1.0000, 0.0031),
        (0.5463, 1.0000, 0.0032),
        (0.2821, 1.0000, 0.0035),
        (0.1500, 1.0000, 0.0036),
        (0.0038, 1.0000, 0.1217),
        (0.0039, 1.0000, 0.2539),
        (0.0039, 1.0000, 0.4901),
        (0.0039, 1.0000, 0.6082),
        (0.0039, 1.0000, 0.8444),
        (0.0039, 1.0000, 0.9625),
        (0.0039, 0.8052, 1.0000),
        (0.0039, 0.6872, 1.0000),
        (0.0040, 0.4510, 1.0000),
        (0.0040, 0.3329, 1.0000),
        (0.0040, 0.0967, 1.0000),
        (0.1474, 0.0040, 1.0000),
        (0.2655, 0.0040, 1.0000),
        (0.5017, 0.0040, 1.0000),
        (0.6198, 0.0040, 1.0000),
        (0.7965, 0.0040, 1.0000),
        (0.8848, 0.0040, 1.0000),
        (1.0000, 0.0040, 0.9424),
        (1.0000, 0.0040, 0.8541),
        (1.0000, 0.0040, 0.6774),
        (1.0000, 0.0040, 0.5890),
        (1.0000, 0.0040, 0.4124),
        (1.0000, 0.0040, 0.3240),
        (1.0000, 0.0040, 0.1473),
        (0.9190, 0.1564, 0.2476),
        (0.7529, 0.3782, 0.4051),
        (0.6699, 0.4477, 0.4584),
        (0.5200, 0.5200, 0.5200),
        (0.4595, 0.4595, 0.4595),
        (0.4100, 0.4100, 0.4100),
        (0.3706, 0.3706, 0.3706),
        (0.2000, 0.2000, 0.2000),
        (0, 0, 0),
    ]

    np_colors_base = np.array(l_colors_dwt)
    np_colors_rgb = np_colors_base[:num_clusters]

    return np_colors_rgb


def colors_fams_3() -> np.ndarray:
    """
    Get colors for 3 wave families.

    Returns
    -------
    np.ndarray
        Array of RGB colors for 3 wave families.
    """

    l_named_colors = [
        "gold",
        "darkgreen",
        "royalblue",
    ]

    np_colors_rgb = np.array([mcolors.to_rgb(c) for c in l_named_colors])

    return np_colors_rgb


def colors_interp(num_clusters: int) -> np.ndarray:
    """
    Generate interpolated colors from Spectral colormap.

    Parameters
    ----------
    num_clusters : int
        Number of clusters to generate colors for.

    Returns
    -------
    np.ndarray
        Array of RGB colors interpolated from Spectral colormap.
    """

    scm = cm.get_cmap("Spectral", num_clusters)
    mnorm = mcolors.Normalize(vmin=0, vmax=num_clusters)
    l_colors = [scm(mnorm(i)) for i in range(num_clusters)]

    return np.array(l_colors)[:, :-1]


def get_cluster_colors(num_clusters: int) -> np.ndarray:
    """
    Get appropriate colors for clustering based on number of clusters.

    Parameters
    ----------
    num_clusters : int
        Number of clusters to get colors for.

    Returns
    -------
    np.ndarray
        Array of RGB colors for the specified number of clusters.
    """

    if num_clusters == 6:
        np_colors_rgb = colors_awt()  # Annual Weather Types
    elif num_clusters == 25:
        np_colors_rgb = colors_mjo()  # MJO Categories
    elif num_clusters in [36, 42]:
        np_colors_rgb = colors_dwt(num_clusters)  # Daily Weather Types
    else:
        np_colors_rgb = colors_interp(num_clusters)  # interpolate

    return np_colors_rgb


def GetFamsColors(num_fams: int) -> np.ndarray:
    """
    Get colors for wave families.

    Parameters
    ----------
    num_fams : int
        Number of wave families.

    Returns
    -------
    np.ndarray
        Array of RGB colors for the specified number of wave families.
    """

    if num_fams == 3:
        np_colors_rgb = colors_fams_3()  # chosen colors
    else:
        np_colors_rgb = colors_interp(num_fams)  # interpolate

    return np_colors_rgb


def colormap_bathy(topat: float, topag: float) -> ListedColormap:
    """
    Create custom colormap for bathymetry plot.

    Parameters
    ----------
    topat : float
        Maximum topography value.
    topag : float
        Minimum bathymetry value.

    Returns
    -------
    ListedColormap
        Custom colormap combining YlGnBu_r and turbid colormaps.
    """

    colors2 = "YlGnBu_r"
    colors1 = cmocean.cm.turbid

    bottom = plt.get_cmap(colors2, -topag * 100)
    top = plt.get_cmap(colors1, topat * 100)

    newcolors = np.vstack(
        (
            bottom(np.linspace(0, 0.8, -topag * 100)),
            top(np.linspace(0.1, 1, topat * 100)),
        )
    )

    return ListedColormap(newcolors)


def colormap_spectra() -> ListedColormap:
    """
    Create custom colormap for spectra plots combining RdBu and rainbow colormaps.

    Returns
    -------
    ListedColormap
        Custom colormap combining RdBu and rainbow colormaps.
    """

    top = cm.get_cmap("RdBu", 128)
    bottom = cm.get_cmap("rainbow", 128)
    newcolors = np.vstack(
        (top(np.linspace(0.5, 0.8, 50)), bottom(np.linspace(0.2, 1, 128)))
    )

    return ListedColormap(newcolors, name="newcmp")



from typing import List, Tuple, Union

import numpy as np
import pandas as pd

from ..core.decorators import validate_data_mda
from ._base_datamining import BaseClustering


def calculate_normalized_squared_distance(
    data_array: Union[np.ndarray, pd.DataFrame],
    array_to_compare: Union[np.ndarray, pd.DataFrame],
    directional_indices: List[int] = None,
    weights: List[float] = None,
) -> np.ndarray:
    """
    Calculate the normalized squared distance between the data_array and the array_to_compare.
    ALERT: directional_indices will be deprecated in the future.

    Parameters
    ----------
    data_array : Union[np.ndarray, pd.DataFrame]
        The data array to compare. Dimensions: (1, n_features).
    array_to_compare : Union[np.ndarray, pd.DataFrame]
        The array to compare against. Dimensions: (n_samples, n_features).
    directional_indices : List[int], optional
        List of column indices that contain directional data.
        For these columns, the minimum circular distance will be used.
        Default is None.
    weights : List[float], optional
        List of weights to apply to each column's distance.
        Must have the same length as the number of columns.
        Default is None (equal weights).

    Returns
    -------
    np.ndarray
        An array of normalized squared distance between the two arrays.
        Dimensions: (n_samples, 1).

    Raises
    ------
    ValueError
        If the arrays have different numbers of columns.
        If weights are provided but length doesn't match number of columns.

    Examples
    --------
    >>> calculate_normalized_squared_distance(
    ...     data_array=np.array([[1, 2, 3]]),
    ...     array_to_compare=np.array([[1, 2, 3], [4, 5, 6]]),
    ... )
    [0.0, 27.0]

    Notes
    -----
    - IMPORTANT: Data is assumed to be normalized before calling this function.
    - For directional variables, the function calculates the minimum circular distance.
      Assuming data is between 0 and 1 (normalized).
    - The function calculates weighted squared differences for each row.
    - If DataFrames are provided, they will be converted to numpy arrays.
    """

    if isinstance(data_array, pd.DataFrame):
        data_array = data_array.values
    if isinstance(array_to_compare, pd.DataFrame):
        array_to_compare = array_to_compare.values

    if data_array.shape[1] != array_to_compare.shape[1]:
        raise ValueError("Arrays must have the same number of columns")

    if weights is not None and len(weights) != data_array.shape[1]:
        raise ValueError("Length of weights must match number of columns")

    # Calculate initial differences
    diff = data_array - array_to_compare

    # Handle directional variables if specified
    if directional_indices is not None:
        for idx in directional_indices:
            # Calculate absolute angular difference
            abs_diff = np.absolute(diff[:, idx])
            # Use minimum circular distance
            diff[:, idx] = np.minimum(abs_diff, 1 - abs_diff)

    # Apply weights if specified
    if weights is not None:
        for i, weight in enumerate(weights):
            diff[:, i] *= weight

    # Compute the squared sum of differences for each row
    dist = np.sum(diff**2, axis=1)

    return dist


def find_nearest_indices(
    query_points: Union[np.ndarray, pd.DataFrame],
    reference_points: Union[np.ndarray, pd.DataFrame],
    directional_indices: List[int] = None,
    weights: List[float] = None,
) -> np.ndarray:
    """
    Find the indices of nearest points in reference_points for each point in query_points.

    Parameters
    ----------
    query_points : Union[np.ndarray, pd.DataFrame]
        The points to find nearest neighbors for.
    reference_points : Union[np.ndarray, pd.DataFrame]
        The set of points to search in.
    directional_indices : List[int], optional
        List of column indices that contain directional data.
        For these columns, the minimum circular distance will be used.
        Default is None.
    weights : List[float], optional
        List of weights to apply to each column's distance.
        Must have the same length as the number of columns.
        Default is None (equal weights).

    Returns
    -------
    np.ndarray
        An array containing the index of the nearest reference point for each query point.

    Examples
    --------
    >>> # Finding nearest centroids for data points
    >>> data = np.random.rand(100, 3)  # 100 points with 3 features
    >>> centroids = np.random.rand(5, 3)  # 5 centroids
    >>> nearest_centroid_indices = find_nearest_indices(data, centroids)
    """

    if isinstance(query_points, pd.DataFrame):
        query_points = query_points.values
    if isinstance(reference_points, pd.DataFrame):
        reference_points = reference_points.values

    nearest_indices = np.zeros(query_points.shape[0], dtype=int)

    for i in range(query_points.shape[0]):
        rep = np.repeat(
            np.expand_dims(query_points[i, :], axis=0),
            reference_points.shape[0],
            axis=0,
        )
        ndist = calculate_normalized_squared_distance(
            data_array=rep,
            array_to_compare=reference_points,
            directional_indices=directional_indices,
            weights=weights,
        )
        nearest_indices[i] = np.nanargmin(ndist)

    return nearest_indices


class MDAError(Exception):
    """
    Custom exception for MDA class.
    """

    def __init__(self, message: str = "MDA error occurred."):
        self.message = message
        super().__init__(self.message)


class MDA(BaseClustering):
    """
    Maximum Dissimilarity Algorithm (MDA) class.

    This class performs the MDA algorithm on a given dataframe.

    Attributes
    ----------
    num_centers : int
        The number of centers to use in the MDA algorithm.
    data_variables : List[str]
        A list with all data variables.
    directional_variables : List[str]
        A list with directional variables.
    fitting_variables : List[str]
        A list with fitting variables.
    custom_scale_factor : dict
        A dictionary of custom scale factors.
    scale_factor : dict
        A dictionary of scale factors (after normalizing the data).
    centroids : pd.DataFrame
        The selected centroids.
    normalized_centroids : pd.DataFrame
        The selected normalized centroids.
    centroid_iterative_indices : List[int]
        A list of iterative indices of the centroids.
    centroid_real_indices : List[int]
        The real indices of the selected centroids.
    is_fitted : bool
        A flag indicating whether the model is fitted or not.

    Examples
    --------
    .. jupyter-execute::

        import numpy as np
        import pandas as pd
        from bluemath_tk.datamining.mda import MDA

        data = pd.DataFrame(
            {
                "Hs": np.random.rand(1000) * 7,
                "Tp": np.random.rand(1000) * 20,
                "Dir": np.random.rand(1000) * 360
            }
        )
        mda = MDA(num_centers=5)
        nearest_centroids_idxs, nearest_centroids_df = mda.fit_predict(
            data=data,
            directional_variables=["Dir"],
        )

        mda.plot_selected_centroids(plot_text=True)
    """

    def __init__(self, num_centers: int) -> None:
        """
        Initializes the MDA class.

        Parameters
        ----------
        num_centers : int
            The number of centers to use in the MDA algorithm.
            Must be greater than 0.

        Raises
        ------
        ValueError
            If num_centers is not greater than 0.
        """

        super().__init__()
        self.set_logger_name(name=self.__class__.__name__)

        if num_centers > 0:
            self.num_centers = int(num_centers)
        else:
            raise ValueError("Variable num_centers must be > 0")

        self._data: pd.DataFrame = pd.DataFrame()
        self._normalized_data: pd.DataFrame = pd.DataFrame()
        self._data_to_fit: pd.DataFrame = pd.DataFrame()
        self.data_variables: List[str] = []
        self.directional_variables: List[str] = []
        self.fitting_variables: List[str] = []
        self.custom_scale_factor: dict = {}
        self.scale_factor: dict = {}
        self.centroids: pd.DataFrame = pd.DataFrame()
        self.normalized_centroids: pd.DataFrame = pd.DataFrame()
        self.centroid_iterative_indices: List[int] = []
        self.centroid_real_indices: np.ndarray = np.array([])
        self.is_fitted: bool = False

    @property
    def data(self) -> pd.DataFrame:
        """
        Returns the original data used for clustering.
        """

        return self._data

    @property
    def normalized_data(self) -> pd.DataFrame:
        """
        Returns the normalized data used for clustering.
        """

        return self._normalized_data

    @property
    def data_to_fit(self) -> pd.DataFrame:
        """
        Returns the data used for fitting the K-Means algorithm.
        """

        return self._data_to_fit

    def _nearest_indices(
        self, normalized_data: pd.DataFrame
    ) -> Tuple[np.ndarray, pd.DataFrame]:
        """
        Compute nearest centroids to the provided data.

        Parameters
        ----------
        normalized_data : pd.DataFrame
            The input data to be used to compute nearest centroids.

        Returns
        -------
        Tuple[np.ndarray, pd.DataFrame]
            An array containing the index of the nearest centroid to the data,
            and a DataFrame containing the nearest centroids.

        Raises
        ------
        MDAError
            If the data is empty.
        """

        if normalized_data.empty:
            raise MDAError("Data cannot be empty.")

        nearest_indices_array = find_nearest_indices(
            query_points=normalized_data,
            reference_points=self.normalized_centroids,
        )

        return nearest_indices_array, self.centroids.iloc[nearest_indices_array]

    @validate_data_mda
    def fit(
        self,
        data: pd.DataFrame,
        directional_variables: List[str] = [],
        custom_scale_factor: dict = {},
        first_centroid_seed: int = None,
        normalize_data: bool = False,
    ) -> None:
        """
        Fit the Maximum Dissimilarity Algorithm (MDA) to the provided data.

        This method initializes centroids for the MDA algorithm using the provided
        dataframe, directional variables, and custom scale factor. It normalizes the
        data, iteratively selects centroids based on maximum dissimilarity, and
        denormalizes the centroids before returning them.

        Parameters
        ----------
        data : pd.DataFrame
            The input data to be used for the MDA algorithm.
        directional_variables : List[str], optional
            A list of directional variables that will be transformed to u and v components.
            Then, to use custom_scale_factor, you must specify the variables names with the u and v suffixes.
            Example: directional_variables=["Dir"], custom_scale_factor={"Dir_u": [0, 1], "Dir_v": [0, 1]}.
            Default is [].
        custom_scale_factor : dict, optional
            A dictionary specifying custom scale factors for normalization.
            If normalize_data is True, this will be used to normalize the data.
            Example: {"Hs": [0, 10], "Tp": [0, 10]}.
            Default is {}.
        first_centroid_seed : int, optional
            The index of the first centroid to use in the MDA algorithm.
            Default is None.
        normalize_data : bool, optional
            A flag to normalize the data.
            If True, the data will be normalized using the custom_scale_factor.
            Default is False.

        Notes
        -----
        - When first_centroid_seed is not provided, max value centroid is used.
        """

        super().fit(
            data=data,
            directional_variables=directional_variables,
            custom_scale_factor=custom_scale_factor,
            normalize_data=normalize_data,
        )

        # Select seed point
        if first_centroid_seed is not None:
            seed = first_centroid_seed
            self.logger.info(f"Using specified seed={seed} as first centroid.")
        else:
            seed = np.argmax(self.normalized_data.sum(axis=1).values)
            self.logger.info(
                f"Using max calculated value seed={seed} as first centroid."
            )

        # Initialize centroids subset
        subset = np.array(
            [self.normalized_data.values[seed]]
        )  # The row that starts as seed
        train = np.delete(self.normalized_data.values, seed, axis=0)

        # Repeat until we have the desired num_centers
        n_c = 1
        while n_c < self.num_centers:
            m2 = subset.shape[0]
            if m2 == 1:
                xx2 = np.repeat(subset, train.shape[0], axis=0)
                d_last = calculate_normalized_squared_distance(
                    data_array=xx2,
                    array_to_compare=train,
                )
            else:
                xx = np.array([subset[-1, :]])
                xx2 = np.repeat(xx, train.shape[0], axis=0)
                d_prev = calculate_normalized_squared_distance(
                    data_array=xx2,
                    array_to_compare=train,
                )
                d_last = np.minimum(d_prev, d_last)

            qerr, bmu = np.nanmax(d_last), np.nanargmax(d_last)

            if not np.isnan(qerr):
                self.centroid_iterative_indices.append(bmu)
                subset = np.append(subset, np.array([train[bmu, :]]), axis=0)
                train = np.delete(train, bmu, axis=0)
                d_last = np.delete(d_last, bmu, axis=0)

                # Log
                fmt = "0{0}d".format(len(str(self.num_centers)))
                self.logger.info(
                    "   MDA centroids: {1:{0}}/{2:{0}}".format(
                        fmt, subset.shape[0], self.num_centers
                    )
                )

            n_c = subset.shape[0]

        # De-normalize scalar and directional data
        self.normalized_centroids = pd.DataFrame(subset, columns=self.fitting_variables)
        self.centroids = self.denormalize(
            normalized_data=self.normalized_centroids, scale_factor=self.scale_factor
        )
        for directional_variable in self.directional_variables:
            self.centroids[directional_variable] = self.get_degrees_from_uv(
                xu=self.centroids[f"{directional_variable}_u"].values,
                xv=self.centroids[f"{directional_variable}_v"].values,
            )

        # Calculate the real indices of the centroids
        self.centroid_real_indices = find_nearest_indices(
            query_points=self.normalized_centroids,
            reference_points=self.normalized_data,
        )

        # Set the fitted flag to True
        self.is_fitted = True

    def predict(self, data: pd.DataFrame) -> Tuple[np.ndarray, pd.DataFrame]:
        """
        Predict the nearest centroid for the provided data.

        Parameters
        ----------
        data : pd.DataFrame
            The input data to be used for the prediction.

        Returns
        -------
        Tuple[np.ndarray, pd.DataFrame]
            A tuple containing the nearest centroid index for each data point and the nearest centroids.
        """

        if self.is_fitted is False:
            raise MDAError("MDA model is not fitted.")

        normalized_data = super().predict(data=data)

        return self._nearest_indices(normalized_data=normalized_data)

    def fit_predict(
        self,
        data: pd.DataFrame,
        directional_variables: List[str] = [],
        custom_scale_factor: dict = {},
        first_centroid_seed: int = None,
        normalize_data: bool = False,
    ) -> Tuple[np.ndarray, pd.DataFrame]:
        """
        Fits the MDA model to the data and predicts the nearest centroids.

        Parameters
        ----------
        data : pd.DataFrame
            The input data to be used for the MDA algorithm.
        directional_variables : List[str], optional
            A list of directional variables that will be transformed to u and v components.
            Then, to use custom_scale_factor, you must specify the variables names with the u and v suffixes.
            Example: directional_variables=["Dir"], custom_scale_factor={"Dir_u": [0, 1], "Dir_v": [0, 1]}.
            Default is [].
        custom_scale_factor : dict, optional
            A dictionary specifying custom scale factors for normalization.
            If normalize_data is True, this will be used to normalize the data.
            Example: {"Hs": [0, 10], "Tp": [0, 10]}.
            Default is {}.
        first_centroid_seed : int, optional
            The index of the first centroid to use in the MDA algorithm.
            Default is None.
        normalize_data : bool, optional
            A flag to normalize the data.
            If True, the data will be normalized using the custom_scale_factor.
            Default is False.

        Returns
        -------
        Tuple[np.ndarray, pd.DataFrame]
            A tuple containing the nearest centroid index for each data point and the nearest centroids.
        """

        self.fit(
            data=data,
            directional_variables=directional_variables,
            custom_scale_factor=custom_scale_factor,
            first_centroid_seed=first_centroid_seed,
            normalize_data=normalize_data,
        )

        return self.predict(data=data)



import numpy as np


def block_maxima(
    x: np.ndarray,
    block_size: int | float = 365.25,
    min_sep: int = 2,
) -> tuple[np.ndarray, np.ndarray]:
    """
    Function to obtain the Block Maxima of given size taking into account
    minimum independence hypothesis

    Parameters
    ----------
    x : np.ndarray
        Data used to compute the Block Maxima
    block_size : int, default=5
        Size of BM in index units (if daily data, nº of days), by default 5
    min_sep : int, optional
        Minimum separation between maximas in index units, by default 2

    Returns
    -------
    idx : np.ndarray
        Indices of BM
    bmaxs : np.ndarray
        BM values

    Raises
    ------
    ValueError
        Minimum separation must be smaller than (block_size+1)/2

    Example
    -------
    >>> # 1-year of daily values
    >>> x = np.random.lognormal(1, 1.2, size=365)

    >>> # 5-day Block Maxima with 72h of independency
    >>> idx, bmaxs = block_maxima(
    >>>     x,
    >>>     block_size=5,
    >>>     min_sep=3
    >>> )
    """
    block_size = int(np.ceil(block_size))

    if min_sep > (block_size + 1) / 2:
        raise ValueError("min_sep must be smaller than (block_size+1)/2")

    x = np.asarray(x)
    n = x.size

    # Partition into non-overlapping blocks
    n_blocks = int(np.ceil(n / block_size))
    segments_idx = []
    segments = []
    blocks = []
    # For each block, keep a *ranked* list of (idx, value) candidates (desc by value)
    for b in range(n_blocks):
        start = b * block_size
        stop = min((b + 1) * block_size, n)
        segment = x[start:stop]

        candidates_idx = np.argsort(segment)[::-1]

        segments_idx.append(np.arange(start, stop))
        segments.append(segment)
        blocks.append(candidates_idx)

    def violates(i_left, i_right):
        if i_left is None or i_right is None:
            return False
        return i_right - i_left < min_sep

    changed = True
    while changed:
        changed = False

        for b in range(n_blocks - 1):
            idx_left = segments_idx[b][blocks[b][0]]
            idx_right = segments_idx[b + 1][blocks[b + 1][0]]
            if not violates(idx_left, idx_right):
                continue
            else:
                idx_block_adjust = (
                    0
                    if segments[b][blocks[b][0]] < segments[b + 1][blocks[b + 1][0]]
                    else 1
                )
                blocks[b + idx_block_adjust] = np.delete(
                    blocks[b + idx_block_adjust], 0
                )
                changed = True
                break

    bmaxs = np.asarray([segments[b][blocks[b][0]] for b in range(n_blocks)])
    idx = np.asarray([segments_idx[b][blocks[b][0]] for b in range(n_blocks)])

    return idx, bmaxs



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# common
from datetime import datetime, date
import os
import os.path as op

# pip
import netCDF4
import numpy as np

# tk
from ..util.time_operations import npdt64todatetime as n2d


# TODO REFACTOR CON teslakit/database.py

def StoreBugXdset(xds_data, p_ncfile):
    '''
    Stores xarray.Dataset to .nc file while avoiding bug with time data (>2262)
    '''

    # get metadata from xarray.Dataset
    dim_names = xds_data.dims.keys()
    var_names = xds_data.variables.keys()

    # Handle time data  (calendar format)
    calendar = 'standard'
    units = 'hours since 1970-01-01 00:00:00'

    # remove previous file
    if op.isfile(p_ncfile):
        os.remove(p_ncfile)

    # Use netCDF4 lib 
    root = netCDF4.Dataset(p_ncfile, 'w', format='NETCDF4')

    # Handle dimensions
    for dn in dim_names:
        vals = xds_data[dn].values[:]
        root.createDimension(dn, len(vals))

    # handle variables
    for vn in var_names:
        vals = xds_data[vn].values[:]

        # dimensions values
        if vn in dim_names:

            if vn == 'time':  # time dimension values
                # TODO: profile / acelerar
                if isinstance(vals[0], datetime):
                    pass

                elif isinstance(vals[0], date):
                    # TODO se pierde la resolucion horaria
                    # parse datetime.date to datetime.datetime
                    vals = [datetime.combine(d, datetime.min.time()) for d in vals]

                elif isinstance(vals[0], np.datetime64):
                    # parse numpy.datetime64 to datetime.datetime
                    vals = [n2d(d) for d in vals]

                dv = root.createVariable(varname=vn, dimensions=(vn,), datatype='int64')
                dv[:] = netCDF4.date2num(vals, units=units, calendar=calendar)
                dv.units = units
                dv.calendar = calendar

            else:
                dv = root.createVariable(varname=vn, dimensions=(vn,), datatype=type(vals[0]))
                dv[:] = vals

        # variables values
        else:
            vdims = xds_data[vn].dims
            vatts = xds_data[vn].attrs

            vv = root.createVariable(varname=vn,dimensions=vdims, datatype='float32')
            vv[:] = vals

            # variable attributes
            vv.setncatts(vatts)

    # global attributes
    root.setncatts(xds_data.attrs)

    # close file
    root.close()





from __future__ import annotations

from abc import abstractmethod
from typing import Dict, List, Tuple

import matplotlib.pyplot as plt
import numpy as np
from scipy.optimize import OptimizeResult, minimize

from ..core.models import BlueMathModel


class BaseDistribution(BlueMathModel):
    """
    Base class for all extreme distributions.
    """

    @abstractmethod
    def __init__(self) -> None:
        """
        Initialize the base distribution class
        """
        super().__init__()

    @staticmethod
    @abstractmethod
    def name() -> str:
        pass

    @staticmethod
    @abstractmethod
    def nparams() -> int:
        pass

    @staticmethod
    @abstractmethod
    def param_names() -> List[str]:
        pass

    @staticmethod
    @abstractmethod
    def pdf(x: np.ndarray, **kwargs) -> np.ndarray:
        """
        Probability density function

        Parameters
        ----------
        x : np.ndarray
            Values to compute the probability density value
        **kwargs :
            Distribution specific parameters as keyword arguments.
            Common parameters include:
            - loc: Location parameter
            - scale: Scale parameter
            - shape: Shape parameter (for some distributions)

        Returns
        ----------
        pdf : np.ndarray
            Probability density function values
        """
        pass

    @staticmethod
    @abstractmethod
    def cdf(x: np.ndarray, **kwargs) -> np.ndarray:
        """
        Cumulative distribution function

        Parameters
        ----------
        x : np.ndarray
            Values to compute their probability
        **kwargs :
            Distribution specific parameters as keyword arguments.
            Common parameters include:
            - loc: Location parameter
            - scale: Scale parameter
            - shape: Shape parameter (for some distributions)

        Returns
        ----------
        p : np.ndarray
            Probability
        """
        pass

    @staticmethod
    @abstractmethod
    def sf(x: np.ndarray, **kwargs) -> np.ndarray:
        """
        Survival function (1-Cumulative Distribution Function)

        Parameters
        ----------
        x : np.ndarray
            Values to compute their survival function value
        **kwargs :
            Distribution specific parameters as keyword arguments.
            Common parameters include:
            - loc: Location parameter
            - scale: Scale parameter
            - shape: Shape parameter (for some distributions)

        Returns
        ----------
        sp : np.ndarray
            Survival function value
        """
        pass

    @staticmethod
    @abstractmethod
    def qf(p: np.ndarray, **kwargs) -> np.ndarray:
        """
        Quantile function

        Parameters
        ----------
        p : np.ndarray
            Probabilities to compute their quantile
        **kwargs :
            Distribution specific parameters as keyword arguments.
            Common parameters include:
            - loc: Location parameter
            - scale: Scale parameter
            - shape: Shape parameter (for some distributions)

        Returns
        ----------
        q : np.ndarray
            Quantile value
        """
        pass

    @staticmethod
    @abstractmethod
    def nll(data: np.ndarray, **kwargs) -> float:
        """
        Negative Log-Likelihood function

        Parameters
        ----------
        data : np.ndarray
            Data to compute the Negative Log-Likelihood value
        **kwargs :
            Distribution specific parameters as keyword arguments.
            Common parameters include:
            - loc: Location parameter
            - scale: Scale parameter
            - shape: Shape parameter (for some distributions)

        Returns
        ----------
        nll : float
            Negative Log-Likelihood value
        """
        pass

    @staticmethod
    @abstractmethod
    def random(size: int, random_state: int, **kwargs) -> np.ndarray:
        """
        Generate random values

        Parameters
        ----------
        size : int
            Number of random values to generate
        random_state : int
            Seed for the random number generator
        **kwargs :
            Distribution specific parameters as keyword arguments.
            Common parameters include:
            - loc: Location parameter
            - scale: Scale parameter
            - shape: Shape parameter (for some distributions)

        Returns
        ----------
        x : np.ndarray
            Random values from GEV distribution
        """
        pass

    @staticmethod
    @abstractmethod
    def mean(**kwargs) -> float:
        """
        Mean

        Parameters
        ----------
        **kwargs :
            Distribution specific parameters as keyword arguments.
            Common parameters include:
            - loc: Location parameter
            - scale: Scale parameter
            - shape: Shape parameter (for some distributions)

        Returns
        ----------
        mean : np.ndarray
            Mean value of GEV with the given parameters
        """
        pass

    @staticmethod
    @abstractmethod
    def median(**kwargs) -> float:
        """
        Median

        Parameters
        ----------
        **kwargs :
            Distribution specific parameters as keyword arguments.
            Common parameters include:
            - loc: Location parameter
            - scale: Scale parameter
            - shape: Shape parameter (for some distributions)

        Returns
        ----------
        median : np.ndarray
            Median value of GEV with the given parameters
        """
        pass

    @staticmethod
    @abstractmethod
    def variance(**kwargs) -> float:
        """
        Variance

        Parameters
        ----------
        **kwargs :
            Distribution specific parameters as keyword arguments.
            Common parameters include:
            - loc: Location parameter
            - scale: Scale parameter
            - shape: Shape parameter (for some distributions)

        Returns
        ----------
        var : np.ndarray
            Variance of GEV with the given parameters
        """
        pass

    @staticmethod
    @abstractmethod
    def std(**kwargs) -> float:
        """
        Standard deviation

        Parameters
        ----------
        **kwargs :
            Distribution specific parameters as keyword arguments.
            Common parameters include:
            - loc: Location parameter
            - scale: Scale parameter
            - shape: Shape parameter (for some distributions)

        Returns
        ----------
        std : np.ndarray
            Standard Deviation of GEV with the given
            parameters
        """
        pass

    @staticmethod
    @abstractmethod
    def stats(**kwargs) -> Dict[str, float]:
        """
        Summary statistics

        Return summary statistics including mean, std, variance, etc.

        Parameters
        ----------
        **kwargs :
            Distribution specific parameters as keyword arguments.
            Common parameters include:
            - loc: Location parameter
            - scale: Scale parameter
            - shape: Shape parameter (for some distributions)

        Returns
        ----------
        stats : dict
            Summary statistics of BaseDistribution with the given
            parameters
        """
        pass

    @staticmethod
    @abstractmethod
    def fit(dist, data: np.ndarray, **kwargs) -> FitResult:
        """
        Fit any distribution

        Parameters
        ----------
        dist : BaseDistribution
            Distribution to fit.
        data : np.ndarray
            Data to fit the GEV distribution
        **kwargs : dict, optional
            Additional keyword arguments for the fitting function.
            These can include options like method, bounds, etc.
            See fit_dist for more details.
            If not provided, default fitting options will be used.

        Returns
        ----------
        FitResult
            Result of the fit containing the parameters loc, scale, shape,
            success status, and negative log-likelihood value.
        """
        pass


class FitResult(BlueMathModel):
    """
    Class used for the results of fitting a distribution

    Attributes
    ----------
    dist : BaseDistribution
        The distribution that was fitted.
    data : np.ndarray
        The data used for fitting the distribution.
    params : np.ndarray
        Fitted parameters of the distribution.
    success : bool
        Indicates whether the fitting was successful.
    message : str
        Message from the optimization result.
    nll : float
        Negative log-likelihood of the fitted distribution.
    res : OptimizeResult
        The result of the optimization process, containing additional information.

    Methods
    -------
    summary() -> dict
        Returns a summary of the fitting results, including parameters, negative log-likelihood,
        success status, message, and the optimization result.
    plot(ax=None, plot_type="hist")
        Plots of fitting results (NOT IMPLEMENTED).

    Notes
    -------
    - This class is used to encapsulate the results of fitting a distribution to data.
    - It provides a method to summarize the fitting results and a placeholder for plotting the results.
    """

    def __init__(self, dist: BaseDistribution, data: np.ndarray, res: OptimizeResult):
        """
        Initialize the FitResult object.

        Parameters
        ----------
        dist : BaseDistribution
            The distribution that was fitted.
        data : np.ndarray
            The data used for fitting the distribution.
        res : OptimizeResult
            The result of the optimization process.
        """
        super().__init__()
        self.dist = dist
        self.data = data

        self.params = res.x
        self.success = res.success
        self.message = res.message
        self.nll = res.fun
        self.res = res

        # Auxiliar for diagnostics plots
        self.n = self.data.shape[0]
        self.ecdf = np.arange(1, self.n + 1) / (self.n + 1)

    def summary(self):
        """
        Print a summary of the fitting results
        """
        print(f"Fitting results for {self.dist.name()}:")
        print("--------------------------------------")
        print("Parameters:")
        for i, param in enumerate(self.params):
            print(f"   - {self.dist.param_names()[i]}: {param:.4f}")
        # print("\n")
        print(f"Negative Log-Likelihood value: {self.nll:.4f}")
        print(f"{self.message}")

    def plot(self, ax: plt.axes = None, plot_type="all") -> Tuple[plt.figure, plt.axes]:
        """
        Plots of fitting results: PP-plot, QQ-plot, histogram with fitted distribution, and return period plot.
        Parameters
        ----------
        ax : matplotlib.axes.Axes, optional
            Axes to plot on. If None, a new figure and axes will be created.
        plot_type : str, optional
            Type of plot to create. Options are "hist" for histogram, "pp" for P-P plot,
            "qq" for Q-Q plot, "return_period" for return period plot, or "all" for all plots.
            Default is "all".

        Returns
        -------
        fig : matplotlib.figure.Figure
            Figure object containing the plots. If `ax` is provided, returns None.

        Raises
        -------
        ValueError
            If `plot_type` is not one of the valid options ("hist", "pp", "qq", "return_period", "all").
        """
        if plot_type == "all":
            fig, axs = plt.subplots(2, 2, figsize=(12, 10))
            self.pp(ax=axs[0, 0])
            self.qq(ax=axs[0, 1])
            self.hist(ax=axs[1, 0])
            self.return_period(ax=axs[1, 1])
            plt.tight_layout()
            return fig, axs
        elif plot_type == "hist":
            return self.hist()
        elif plot_type == "pp":
            return self.pp()
        elif plot_type == "qq":
            return self.qq()
        elif plot_type == "return_period":
            return self.return_period()
        else:
            raise ValueError(
                "Invalid plot type. Use 'hist', 'pp', 'qq', 'return_period', or 'all'."
            )

    def pp(self, ax: plt.axes = None) -> Tuple[plt.figure, plt.axes]:
        """
        Probability plot of the fitted distribution.

        Parameters
        ----------
        ax : matplotlib.axes.Axes, optional
            Axes to plot on. If None, a new figure and axes will be created.
        """
        if ax is None:
            fig, ax = plt.subplots(figsize=(8, 6))
        else:
            fig = None

        probabilities = self.dist.cdf(np.sort(self.data), *self.params)
        ax.plot([0, 1], [0, 1], color="tab:red", linestyle="--")
        ax.plot(
            probabilities,
            self.ecdf,
            color="tab:blue",
            marker="o",
            linestyle="",
            alpha=0.7,
        )
        ax.set_xlabel("Fitted Probability")
        ax.set_ylabel("Empirical Probability")
        ax.set_title("PP Plot")
        ax.grid()

        return fig, ax

    def qq(self, ax: plt.axes = None) -> Tuple[plt.figure, plt.axes]:
        """
        Quantile-Quantile plot of the fitted distribution.

        Parameters
        ----------
        ax : matplotlib.axes.Axes, optional
            Axes to plot on. If None, a new figure and axes will be created.
        """
        if ax is None:
            fig, ax = plt.subplots(figsize=(8, 6))
        else:
            fig = None

        quantiles = self.dist.qf(self.ecdf, *self.params)
        ax.plot(
            [np.min(self.data), np.max(self.data)],
            [np.min(self.data), np.max(self.data)],
            color="tab:red",
            linestyle="--",
        )
        ax.plot(
            quantiles,
            np.sort(self.data),
            color="tab:blue",
            marker="o",
            linestyle="",
            alpha=0.7,
        )
        ax.set_xlabel("Theoretical Quantiles")
        ax.set_ylabel("Sample Quantiles")
        ax.set_title("QQ Plot")
        ax.grid()

        return fig, ax

    def hist(self, ax: plt.axes = None) -> Tuple[plt.figure, plt.axes]:
        """
        Histogram of the data with the fitted distribution overlayed.

        Parameters
        ----------
        ax : matplotlib.axes.Axes, optional
            Axes to plot on. If None, a new figure and axes will be created.
        """
        if ax is None:
            fig, ax = plt.subplots(figsize=(8, 6))
        else:
            fig = None

        ax.hist(
            self.data,
            bins=30,
            density=True,
            alpha=0.7,
            color="tab:blue",
            label="Data Histogram",
        )
        x = np.linspace(np.min(self.data), np.max(self.data), 1000)
        ax.plot(x, self.dist.pdf(x, *self.params), color="tab:red", label="Fitted PDF")
        ax.set_xlabel("Data Values")
        ax.set_ylabel("Density")
        ax.set_title("Histogram and Fitted PDF")
        ax.legend()
        ax.grid()

        return fig, ax

    def return_period(self, ax: plt.axes = None) -> Tuple[plt.figure, plt.axes]:
        """
        Return period plot of the fitted distribution.

        Parameters
        ----------
        ax : matplotlib.axes.Axes, optional
            Axes to plot on. If None, a new figure and axes will be created.
        """
        if ax is None:
            fig, ax = plt.subplots(figsize=(8, 6))
        else:
            fig = None

        return_years = np.asarray([1.001, 1.1, 1.2, 1.3, 1.4, 1.5, 1.75, 2, 3, 4, 5, 7.5, 10, 15, 20, 25, 50, 100, 250, 500, 1000])
        ecdf_fitted = 1 - 1/return_years
        sorted_data = np.sort(self.data)
        exceedance_prob = 1 - self.ecdf
        return_period = 1 / exceedance_prob

        ax.plot(
            return_years,
            self.dist.qf(ecdf_fitted, *self.params),
            color="tab:red",
            label="Fitted Distribution",
        )
        ax.plot(
            return_period,
            sorted_data,
            marker="o",
            linestyle="",
            color="tab:blue",
            alpha=0.7,
            label="Empirical Data",
        )
        ax.set_xscale("log")
        ax.set_xticks([1, 2, 5, 10, 25, 50, 100, 250, 1000, 10000])
        ax.set_xticklabels([1, 2, 5, 10, 25, 50, 100, 500, 1000, 10000])
        # ax.set_xlim(right=np.max(return_period) * 1.2)
        ax.set_xlabel("Return Period")
        ax.set_ylabel("Data Values")
        ax.set_title("Return Period Plot")
        ax.legend()
        ax.grid()

        return fig, ax


def fit_dist(dist: BaseDistribution, data: np.ndarray, **kwargs) -> FitResult:
    """
    Fit a distribution to data using Maximum Likelihood Estimation (MLE).

    Parameters
    ----------
    dist : BaseDistribution
        Distribution to fit.
    data : np.ndarray
        Data to use for fitting the distribution.
    **kwargs : dict, optional
        Additional options for fitting:
        - 'x0': Initial guess for distribution parameters (default: [mean, std, 0.0]).
        - 'method': Optimization method (default: 'Nelder-Mead').
        - 'bounds': Bounds for optimization parameters (default: [(None, None), (0, None), ...]).
        - 'options': Options for the optimizer (default: {'disp': False}).
        - 'f0', 'floc': Fix location parameter (first parameter).
        - 'f1', 'fscale': Fix scale parameter (second parameter).
        - 'f2', 'fshape': Fix shape parameter (third parameter).

    Returns
    -------
    FitResult
        The fitting results, including parameters, success status, and negative log-likelihood.
    """
    nparams = dist.nparams()
    param_names = dist.param_names()

    # Handle fixed parameters
    fixed_params = {}
    for i, name in enumerate(param_names):
        # Check both numerical (f0, f1, f2) and named (floc, fscale, fshape) fixed parameters
        num_key = f"f{i}"
        if num_key in kwargs:
            fixed_params[i] = kwargs[num_key]

    # Create list of free parameters (those not fixed)
    free_params = [i for i in range(nparams) if i not in fixed_params]
    # n_free = len(free_params)

    # Default optimization settings for free parameters only
    default_x0 = np.asarray([np.mean(data), np.std(data)] + [0.1] * (nparams - 2))
    x0 = kwargs.get("x0", default_x0)
    x0_free = np.array([x0[i] for i in free_params])

    method = kwargs.get("method", "Nelder-Mead").lower()
    default_bounds = [(None, None), (1e-6, None)] + [(None, None)] * (nparams - 2)
    bounds = kwargs.get("bounds", default_bounds)
    bounds_free = [bounds[i] for i in free_params]
    options = kwargs.get("options", {"disp": False})

    # Objective function that handles fixed parameters
    def obj(free_values):
        # Reconstruct full parameter vector with fixed values
        full_params = np.zeros(nparams)
        free_idx = 0
        for i in range(nparams):
            if i in fixed_params:
                full_params[i] = fixed_params[i]
            else:
                full_params[i] = free_values[free_idx]
                free_idx += 1
        return dist.nll(data, *full_params)

    # Perform optimization only on free parameters
    result = minimize(
        fun=obj, x0=x0_free, method=method, bounds=bounds_free, options=options
    )

    # Reconstruct full parameter vector for final result
    full_params = np.zeros(nparams)
    free_idx = 0
    for i in range(nparams):
        if i in fixed_params:
            full_params[i] = fixed_params[i]
        else:
            full_params[i] = result.x[free_idx]
            free_idx += 1

    # Create modified result with full parameters
    # TODO: Modify the OptimizeResult to include all the attributes from the original result
    # but with the full parameter vector
    # TODO: Add hess info or hess_inv info to compute confidence intervals in FitResult
    modified_result = OptimizeResult(
        x=full_params,
        success=result.success,
        fun=result.fun,
        message=result.message,
        hess_inv=result.hess_inv if hasattr(result, "hess_inv") else None,
    )

    # Return the fitting result as a FitResult object
    return FitResult(dist, data, modified_result)



from typing import Tuple, Union

import cartopy.crs as ccrs
import cartopy.feature as cfeature
import matplotlib as mpl
import numpy as np
import pandas as pd
import statsmodels.api as sm
import xarray as xr
from matplotlib import pyplot as plt
from matplotlib.axes import Axes
from matplotlib.figure import Figure

from ..core.decorators import validate_data_calval
from ..core.models import BlueMathModel
from ..core.plotting.scatter import density_scatter, validation_scatter


def get_matching_times_between_arrays(
    times1: np.ndarray,
    times2: np.ndarray,
    max_time_diff: int,
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Finds matching time indices between two arrays of timestamps.

    For each time in `times1`, finds the closest time in `times2` that is within `max_time_diff` hours.
    Returns the indices of matching times in both arrays.

    Parameters
    ----------
    times1 : np.ndarray
        First array of timestamps (reference times, e.g., from model data).
    times2 : np.ndarray
        Second array of timestamps (e.g., from satellite or validation data).
    max_time_diff : int
        Maximum time difference in hours for considering times as matching.

    Returns
    -------
    Tuple[np.ndarray, np.ndarray]
        Two arrays containing the indices of matching times:
        - First array: indices in times1 that have matches
        - Second array: corresponding indices in times2 that match

    Example
    -------
    >>> idx1, idx2 = get_matching_times_between_arrays(
    ...     model_df.index.values,
    ...     sat_df.index.values,
    ...     max_time_diff=2,
    ... )
    """

    indices1 = np.array([], dtype=int)
    indices2 = np.array([], dtype=int)

    for i in range(len(times1)):
        # Find minimum time difference for current time1
        time_diffs = np.abs(times2 - times1[i])
        min_diff = np.min(time_diffs)

        # If minimum difference is within threshold, record the indices
        if min_diff < np.timedelta64(max_time_diff, "h"):
            min_index = np.argmin(time_diffs)
            indices1 = np.append(indices1, i)
            indices2 = np.append(indices2, min_index)

    return indices1, indices2


def process_imos_satellite_data(
    satellite_df: pd.DataFrame,
    ini_lat: float,
    end_lat: float,
    ini_lon: float,
    end_lon: float,
    depth_threshold: float = -200,
) -> pd.DataFrame:
    """
    Processes IMOS satellite data for calibration.

    This function filters and processes IMOS satellite altimeter data to be used as
    reference data for calibration (e.g., as `data_to_calibrate` in CalVal.fit).

    Parameters
    ----------
    satellite_df : pd.DataFrame
        IMOS satellite data. Must contain columns:
            - 'LATITUDE' (float): Latitude in decimal degrees
            - 'LONGITUDE' (float): Longitude in decimal degrees
            - 'SWH_KU_quality_control' (float): Quality control flag for Ku-band
            - 'SWH_KA_quality_control' (float): Quality control flag for Ka-band
            - 'SWH_KU_CAL' (float): Calibrated significant wave height (Ku-band)
            - 'SWH_KA_CAL' (float): Calibrated significant wave height (Ka-band)
            - 'BOT_DEPTH' (float): Bathymetry (negative values for ocean)
    ini_lat : float
        Minimum latitude (southern boundary) for filtering.
    end_lat : float
        Maximum latitude (northern boundary) for filtering.
    ini_lon : float
        Minimum longitude (western boundary) for filtering.
    end_lon : float
        Maximum longitude (eastern boundary) for filtering.
    depth_threshold : float, optional
        Only include points with BOT_DEPTH < depth_threshold. Default is -200.

    Returns
    -------
    pd.DataFrame
        Filtered and processed satellite data, suitable for use as `data_to_calibrate` in CalVal.fit.
        Includes a new column 'Hs_CAL' (combination of Ku-band and Ka-band calibrated significant wave heights).

    Notes
    -----
    - The returned DataFrame can be used directly as the `data_to_calibrate` argument in CalVal.fit.
    """

    # Filter satellite data by coordinates
    satellite_df = satellite_df[
        (satellite_df.LATITUDE > ini_lat)
        & (satellite_df.LATITUDE < end_lat)
        & (satellite_df.LONGITUDE > ini_lon)
        & (satellite_df.LONGITUDE < end_lon)
        & (satellite_df.BOT_DEPTH < depth_threshold)
    ]

    # Process quality control
    wave_height_qlt = np.nansum(
        np.concatenate(
            (
                satellite_df["SWH_KU_quality_control"].values[:, np.newaxis],
                satellite_df["SWH_KA_quality_control"].values[:, np.newaxis],
            ),
            axis=1,
        ),
        axis=1,
    )
    good_qlt = np.where(wave_height_qlt < 1.5)

    # Process wave heights
    satellite_df["Hs_CAL"] = np.nansum(
        np.concatenate(
            (
                satellite_df["SWH_KU_CAL"].values[:, np.newaxis],
                satellite_df["SWH_KA_CAL"].values[:, np.newaxis],
            ),
            axis=1,
        ),
        axis=1,
    )

    return satellite_df.iloc[good_qlt]


class CalVal(BlueMathModel):
    """
    Calibrates wave data using reference data.

    This class provides a framework for calibrating wave model outputs (e.g., hindcast or reanalysis)
    using reference data (e.g., satellite or buoy observations).
    It supports directionally-dependent calibration for both sea and swell components.

    Attributes
    ----------
    direction_bin_size : int
        Size of directional bins in degrees.
    direction_bins : np.ndarray
        Array of bin edges for directions.
    calibration_model : sm.OLS
        The calibration model, more details in `statsmodels.api.OLS`.
    calibrated_data : pd.DataFrame
        DataFrame with columns ['Hs', 'Hs_CORR', 'Hs_CAL'] after calibration.
        The time domain is the same as the model data.
    calibration_params : dict
        Dictionary with 'sea_correction' and 'swell_correction' correction coefficients.

    Examples
    --------
    .. jupyter-execute::

        import pandas as pd
        from bluemath_tk.waves.calibration import CalVal, process_imos_satellite_data

        # Load your model data (must have columns: 'Hs', 'Hsea', 'Dirsea', 'Hswell1', 'Dirswell1', ...)
        model_df = pd.read_csv('model_data.csv', index_col=0, parse_dates=True)

        # Load IMOS satellite data and process it for calibration
        sat_df = pd.read_csv('imos_satellite.csv', index_col=0, parse_dates=True)
        data_to_calibrate = process_imos_satellite_data(
            sat_df, ini_lat=-40, end_lat=-30, ini_lon=140, end_lon=150
        )

        # Initialize and fit the calibration
        cal = CalVal()
        cal.fit(
            data=model_df,
            data_longitude=145.0,
            data_latitude=-35.0,
            data_to_calibrate=data_to_calibrate,
            max_time_diff=2,
        )

        # Plot results
        cal.plot_calibration_results()

    """

    direction_bin_size: int = 22.5
    direction_bins: np.ndarray = np.arange(
        direction_bin_size, 360.5, direction_bin_size
    )

    def __init__(self) -> None:
        """
        Initialize the CalVal class.
        """

        super().__init__()
        self.set_logger_name(name="CalVal", level="INFO", console=True)

        # Save input data
        self._data: pd.DataFrame = None
        self._data_longitude: float = None
        self._data_latitude: float = None
        self._data_to_calibrate: pd.DataFrame = None
        self._max_time_diff: int = None

        # Initialize calibration results
        self._data_to_fit: Tuple[pd.DataFrame, pd.DataFrame] = (None, None)
        self._calibration_model: sm.OLS = None
        self._calibrated_data: pd.DataFrame = None
        self._calibration_params: pd.Series = None

        # Exclude large attributes from model saving
        self._exclude_attributes += [
            "_data",
            "_data_to_calibrate",
            "_data_to_fit",
        ]

    @property
    def calibration_model(self) -> sm.OLS:
        """Returns the calibration model."""

        if self._calibration_model is None:
            raise ValueError(
                "Calibration model is not available. Please run the fit method first."
            )

        return self._calibration_model

    @property
    def calibrated_data(self) -> pd.DataFrame:
        """Returns the calibrated data."""

        if self._calibrated_data is None:
            raise ValueError(
                "Calibrated data is not available. Please run the fit method first."
            )

        return self._calibrated_data

    @property
    def calibration_params(self) -> pd.Series:
        """Returns the calibration parameters."""

        if self._calibration_params is None:
            raise ValueError(
                "Calibration parameters are not available. Please run the fit method first."
            )

        return self._calibration_params

    def _plot_data_domains(self) -> Tuple[Figure, Axes]:
        """
        Plots the domains of the data points.

        Returns
        -------
        Tuple[Figure, Axes]
            A tuple containing the figure and axes objects.
        """

        fig, ax = plt.subplots(
            figsize=(10, 10),
            subplot_kw={
                "projection": ccrs.PlateCarree(central_longitude=self._data_longitude)
            },
        )
        land_10m = cfeature.NaturalEarthFeature(
            "physical",
            "land",
            "10m",
            edgecolor="face",
            facecolor=cfeature.COLORS["land"],
        )
        # Plot calibration data
        ax.scatter(
            self._data_to_calibrate.LONGITUDE,
            self._data_to_calibrate.LATITUDE,
            s=0.01,
            c="k",
            transform=ccrs.PlateCarree(),
        )
        # Plot main data point
        ax.scatter(
            self._data_longitude,
            self._data_latitude,
            s=50,
            c="red",
            zorder=10,
            transform=ccrs.PlateCarree(),
        )
        # Set plot extent
        ax.set_extent(
            [
                self._data_longitude - 2,
                self._data_longitude + 2,
                self._data_latitude - 2,
                self._data_latitude + 2,
            ]
        )
        ax.set_facecolor("lightblue")
        ax.add_feature(land_10m)

        return fig, ax

    def _create_vec_direc(self, waves: np.ndarray, direcs: np.ndarray) -> np.ndarray:
        """
        Creates a vector of wave heights for each directional bin.

        Parameters
        ----------
        waves : np.ndarray
            Wave heights.
        direcs : np.ndarray
            Wave directions in degrees.

        Returns
        -------
        np.ndarray
            Matrix of wave heights for each directional bin.
        """

        data = np.zeros((len(waves), len(self.direction_bins)))
        for i in range(len(waves)):
            if direcs[i] < 0:
                direcs[i] = direcs[i] + 360
            if direcs[i] > 0 and waves[i] > 0:
                # Handle direction = 360° case by mapping to the first bin (0-22.5°)
                if direcs[i] >= 360:
                    bin_idx = 0
                else:
                    bin_idx = int(direcs[i] / self.direction_bin_size)
                data[i, bin_idx] = waves[i]

        return data

    @staticmethod
    def _get_nparts(data: pd.DataFrame) -> int:
        """
        Gets the number of parts in the wave data.

        Parameters
        ----------
        data : pd.DataFrame
            Wave data.

        Returns
        -------
        int
            The number of parts in the wave data.
        """

        return len([col for col in data.columns if col.startswith("Hswell")])

    def _get_joined_sea_swell_data(self, data: pd.DataFrame) -> np.ndarray:
        """
        Joins the sea and swell data.

        Parameters
        ----------
        data : pd.DataFrame
            Wave data.

        Returns
        -------
        np.ndarray
            The joined sea and swell matrix.
        """

        # Process sea waves
        Hsea = self._create_vec_direc(data["Hsea"], data["Dirsea"]) ** 2

        # Process swells
        Hs_swells = np.zeros(Hsea.shape)
        for part in range(1, self._get_nparts(data) + 1):
            Hs_swells += (
                self._create_vec_direc(data[f"Hswell{part}"], data[f"Dirswell{part}"])
            ) ** 2

        # Combine sea and swell matrices
        sea_swell_matrix = np.concatenate([Hsea, Hs_swells], axis=1)

        return sea_swell_matrix

    @validate_data_calval
    def fit(
        self,
        data: pd.DataFrame,
        data_longitude: float,
        data_latitude: float,
        data_to_calibrate: pd.DataFrame,
        max_time_diff: int = 2,
    ) -> None:
        """
        Calibrate the model data using reference (calibration) data.

        This method matches the model data and calibration data in time,
        constructs directionally-binned sea and swell matrices,
        and fits a linear regression to obtain correction coefficients
        for each direction bin.

        Parameters
        ----------
        data : pd.DataFrame
            Model data to calibrate. Must contain columns:
                - 'Hs' (float): Significant wave height
                - 'Hsea' (float): Sea component significant wave height
                - 'Dirsea' (float): Sea component mean direction (degrees)
                - 'Hswell1', 'Dirswell1', ... (float): Swell components (at least one required)
            The index must be datetime-like.
        data_longitude : float
            Longitude of the model location (used for plotting and filtering).
        data_latitude : float
            Latitude of the model location (used for plotting and filtering).
        data_to_calibrate : pd.DataFrame
            Reference data for calibration. Must contain column:
                - 'Hs_CAL' (float): Calibrated significant wave height (e.g., from satellite)
            The index must be datetime-like.
        max_time_diff : int, optional
            Maximum time difference (in hours) allowed when matching model and calibration data.
            Default is 2.

        Notes
        -----
        - After calling this method, the calibration parameters are stored in `self.calibration_params`
        and the calibrated data is available in `self.calibrated_data`.
        - The calibration is directionally dependent, meaning it uses different correction coefficients
        for different wave directions.
        - The coefficients with p-values greater than 0.05 or negative values are set to 1.0,
        indicating no correction is applied for those directions.
        """

        self.logger.info("Starting calibration fit procedure.")

        # Save input data
        self._data = data.copy()
        self._data_longitude = data_longitude
        self._data_latitude = data_latitude
        self._data_to_calibrate = data_to_calibrate.copy()
        self._max_time_diff = max_time_diff

        # Plot data domains
        self.logger.info("Plotting data domains.")
        self._plot_data_domains()

        # Construct matrices for calibration
        self.logger.info("Matching times and constructing matrices for calibration.")

        # Get matching times
        times_data_to_fit, times_data_to_calibrate = get_matching_times_between_arrays(
            self._data.index.values,
            self._data_to_calibrate.index.values,
            max_time_diff=self._max_time_diff,
        )
        self._data_to_fit = (
            self._data.iloc[times_data_to_fit],
            self._data_to_calibrate.iloc[times_data_to_calibrate],
        )

        # Get joined sea and swell data
        sea_swell_matrix = self._get_joined_sea_swell_data(self._data_to_fit[0])

        # Perform calibration
        self.logger.info("Fitting OLS regression for calibration.")
        X = sm.add_constant(sea_swell_matrix)
        self._calibration_model = sm.OLS(self._data_to_fit[1]["Hs_CAL"] ** 2, X)
        calibrated_model_results = self._calibration_model.fit()

        # Get significant correction coefficients
        significant_model_params = [
            model_param
            if calibrated_model_results.pvalues[imp] < 0.05 and model_param > 0
            else 1.0
            for imp, model_param in enumerate(calibrated_model_results.params)
        ]

        # Save sea and swell correction coefficients
        self._calibration_params = {
            "sea_correction": {
                ip: param
                for ip, param in enumerate(
                    np.sqrt(significant_model_params[: len(self.direction_bins)])
                )
            },
            "swell_correction": {
                ip: param
                for ip, param in enumerate(
                    np.sqrt(significant_model_params[len(self.direction_bins) :])
                )
            },
        }

        # Save calibrated data to be used in plot_calibration_results()
        self._calibrated_data = self.correct(self._data_to_fit[0])
        self._calibrated_data["Hs_CAL"] = self._data_to_fit[1]["Hs_CAL"].values

        self.logger.info("Calibration fit procedure completed.")

    def correct(
        self, data: Union[pd.DataFrame, xr.Dataset]
    ) -> Union[pd.DataFrame, xr.Dataset]:
        """
        Apply the calibration correction to new data.

        Parameters
        ----------
        data : pd.DataFrame or xr.Dataset
            Data to correct. If DataFrame, must contain columns:
                - 'Hs', 'Hsea', 'Dirsea', 'Hswell1', 'Dirswell1', ...
            If xarray.Dataset, must have variable 'efth' and dimension 'part'.

        Returns
        -------
        pd.DataFrame or xr.Dataset
            Corrected data. For DataFrame, returns columns ['Hs', 'Hs_CORR'] (original and corrected SWH).
            For Dataset, adds variables 'corr_coeffs' and 'corr_efth'.

        Notes
        -----
        - The correction is directionally dependent and uses the coefficients obtained from `fit`.
        """

        if self._calibration_params is None:
            raise ValueError(
                "Calibration parameters are not available. Run fit() first."
            )

        if isinstance(data, xr.Dataset):
            self.logger.info(
                "Input is xarray.Dataset. Applying correction to spectra data."
            )

            corrected_data = data.copy()  # Copy data to avoid modifying original data
            peak_directions = corrected_data.spec.stats(["dp"]).load()
            correction_coeffs = np.ones(peak_directions.dp.shape)
            for n_part in peak_directions.part:
                if n_part == 0:
                    correction_coeffs[n_part, :] = np.array(
                        [
                            self.calibration_params["sea_correction"][
                                int(peak_direction / self.direction_bin_size)
                                if peak_direction < 360
                                else 0  # TODO: Check if this with Javi
                            ]
                            for peak_direction in peak_directions.isel(
                                part=n_part
                            ).dp.values
                        ]
                    )
                else:
                    correction_coeffs[n_part, :] = np.array(
                        [
                            self.calibration_params["swell_correction"][
                                int(peak_direction / self.direction_bin_size)
                                if peak_direction < 360
                                else 0  # TODO: Check if this with Javi
                            ]
                            for peak_direction in peak_directions.isel(
                                part=n_part
                            ).dp.values
                        ]
                    )
            corrected_data["corr_coeffs"] = (("part", "time"), correction_coeffs)
            corrected_data["corr_efth"] = (
                corrected_data.efth * corrected_data.corr_coeffs
            )
            self.logger.info("Spectra correction complete.")

            return corrected_data

        elif isinstance(data, pd.DataFrame):
            self.logger.info(
                "Input is pandas.DataFrame. Applying correction to wave data."
            )

            corrected_data = data.copy()
            corrected_data["Hsea"] = (
                corrected_data["Hsea"] ** 2
                * np.array(
                    [
                        self.calibration_params["sea_correction"][
                            int(peak_direction / self.direction_bin_size)
                            if peak_direction < 360
                            else 0
                        ]
                        for peak_direction in corrected_data["Dirsea"]
                    ]
                )
                ** 2
            )
            corrected_data["Hs_CORR"] = corrected_data["Hsea"]
            for n_part in range(1, self._get_nparts(corrected_data) + 1):
                corrected_data[f"Hswell{n_part}"] = (
                    corrected_data[f"Hswell{n_part}"] ** 2
                    * np.array(
                        [
                            self.calibration_params["swell_correction"][
                                int(peak_direction / self.direction_bin_size)
                                if peak_direction < 360
                                else 0
                            ]
                            for peak_direction in corrected_data[f"Dirswell{n_part}"]
                        ]
                    )
                    ** 2
                )
                corrected_data["Hs_CORR"] += corrected_data[f"Hswell{n_part}"]

            corrected_data["Hs_CORR"] = np.sqrt(corrected_data["Hs_CORR"])
            self.logger.info("Wave data correction complete.")

            return corrected_data[["Hs", "Hs_CORR"]]

    def plot_calibration_results(self) -> Tuple[Figure, list]:
        """
        Plot the calibration results, including:
        - Pie charts of correction coefficients for sea and swell
        - Scatter plots of model vs. reference (before and after correction)
        - Polar density plots of sea and swell wave climate

        Returns
        -------
        Tuple[Figure, list]
            The matplotlib Figure and a list of Axes objects for all subplots.
        """

        self.logger.info("Plotting calibration results.")

        fig = plt.figure(figsize=(10, 15))
        gs = fig.add_gridspec(8, 2, wspace=0.4, hspace=0.7)

        # Create subplots with proper projections
        ax1 = fig.add_subplot(gs[:2, 0])  # Sea correction pie
        ax2 = fig.add_subplot(gs[:2, 1])  # Swell correction pie
        ax1_cbar = fig.add_subplot(gs[2, 0])  # Sea correction colorbar
        ax2_cbar = fig.add_subplot(gs[2, 1])  # Swell correction colorbar
        ax3 = fig.add_subplot(gs[3:5, 0])  # No correction scatter
        ax4 = fig.add_subplot(gs[3:5, 1])  # With correction scatter
        ax5 = fig.add_subplot(gs[6:, 0], projection="polar")  # Sea climate
        ax6 = fig.add_subplot(gs[6:, 1], projection="polar")  # Swell climate

        # Plot sea correction pie chart
        sea_norm = 0.35  # Normalization factor for sea correction
        sea_fracs = np.repeat(10, len(self.calibration_params["sea_correction"]))
        sea_norm = mpl.colors.Normalize(1 - sea_norm, 1 + sea_norm)
        sea_cmap = mpl.cm.get_cmap(
            "bwr", len(self.calibration_params["sea_correction"])
        )
        sea_colors = sea_cmap(
            sea_norm(list(self.calibration_params["sea_correction"].values()))
        )
        ax1.pie(
            sea_fracs,
            labels=None,
            colors=sea_colors,
            startangle=90,
            counterclock=False,
            radius=1.2,
        )
        ax1.set_title("SEA $Correction$", fontweight="bold")
        # Add colorbar for sea correction below the pie chart, shrink it
        _sea_cbar = mpl.colorbar.ColorbarBase(
            ax1_cbar,
            cmap=sea_cmap,
            norm=sea_norm,
            orientation="horizontal",
            label="Correction Factor",
        )
        box = ax1_cbar.get_position()
        ax1_cbar.set_position(
            [
                box.x0 + 0.15 * box.width,
                box.y0 + 0.3 * box.height,
                0.7 * box.width,
                0.4 * box.height,
            ]
        )
        ax1_cbar.set_frame_on(False)
        ax1_cbar.tick_params(
            left=False, right=False, labelleft=False, labelbottom=True, bottom=True
        )

        # Plot swell correction pie chart
        swell_norm = 0.35  # Normalization factor for swell correction
        swell_fracs = np.repeat(10, len(self.calibration_params["swell_correction"]))
        swell_norm = mpl.colors.Normalize(1 - swell_norm, 1 + swell_norm)
        swell_cmap = mpl.cm.get_cmap(
            "bwr", len(self.calibration_params["swell_correction"])
        )
        swell_colors = swell_cmap(
            swell_norm(list(self.calibration_params["swell_correction"].values()))
        )
        ax2.pie(
            swell_fracs,
            labels=None,
            colors=swell_colors,
            startangle=90,
            counterclock=False,
            radius=1.2,
        )
        ax2.set_title("SWELL $Correction$", fontweight="bold")
        # Add colorbar for swell correction below the pie chart, shrink it
        _swell_cbar = mpl.colorbar.ColorbarBase(
            ax2_cbar,
            cmap=swell_cmap,
            norm=swell_norm,
            orientation="horizontal",
            label="Correction Factor",
        )
        box = ax2_cbar.get_position()
        ax2_cbar.set_position(
            [
                box.x0 + 0.15 * box.width,
                box.y0 + 0.3 * box.height,
                0.7 * box.width,
                0.4 * box.height,
            ]
        )
        ax2_cbar.set_frame_on(False)
        ax2_cbar.tick_params(
            left=False, right=False, labelleft=False, labelbottom=True, bottom=True
        )

        # Plot no correction scatter
        validation_scatter(
            axs=ax3,
            x=self.calibrated_data["Hs"].values,
            y=self.calibrated_data["Hs_CAL"].values,
            xlabel="Hindcast",
            ylabel="Satellite",
            title="No Correction",
        )

        # Plot with correction scatter
        validation_scatter(
            axs=ax4,
            x=self.calibrated_data["Hs_CORR"].values,
            y=self.calibrated_data["Hs_CAL"].values,
            xlabel="Hindcast",
            ylabel="Satellite",
            title="With Correction",
        )

        # Plot sea wave climate
        sea_dirs = self._data["Dirsea"].iloc[::10] * np.pi / 180
        sea_heights = self._data["Hsea"].iloc[::10]
        # Filter out NaN and infinite values
        valid_mask = np.isfinite(sea_dirs) & np.isfinite(sea_heights)
        sea_dirs_valid = sea_dirs[valid_mask]
        sea_heights_valid = sea_heights[valid_mask]

        if len(sea_dirs_valid) > 0:
            x, y, z = density_scatter(sea_dirs_valid, sea_heights_valid)
            ax5.scatter(x, y, c=z, s=3, cmap="jet")
        ax5.set_theta_zero_location("N", offset=0)
        ax5.set_xticklabels(["N", "NE", "E", "SE", "S", "SW", "W", "NW"])
        ax5.xaxis.grid(True, color="lavender", linestyle="-")
        ax5.yaxis.grid(True, color="lavender", linestyle="-")
        ax5.set_theta_direction(-1)
        ax5.set_xlabel("$\u03b8_{m}$ ($\degree$)")
        ax5.set_ylabel("$H_{s}$ (m)", labelpad=20)
        ax5.set_title("SEA $Wave$ $Climate$", pad=35, fontweight="bold")

        # Plot swell wave climate
        swell_dirs = self._data["Dirswell1"].iloc[::10] * np.pi / 180
        swell_heights = self._data["Hswell1"].iloc[::10]
        # Filter out NaN and infinite values
        valid_mask = np.isfinite(swell_dirs) & np.isfinite(swell_heights)
        swell_dirs_valid = swell_dirs[valid_mask]
        swell_heights_valid = swell_heights[valid_mask]

        if len(swell_dirs_valid) > 0:
            x, y, z = density_scatter(swell_dirs_valid, swell_heights_valid)
            ax6.scatter(x, y, c=z, s=3, cmap="jet")
        ax6.set_theta_zero_location("N", offset=0)
        ax6.set_xticklabels(["N", "NE", "E", "SE", "S", "SW", "W", "NW"])
        ax6.xaxis.grid(True, color="lavender", linestyle="-")
        ax6.yaxis.grid(True, color="lavender", linestyle="-")
        ax6.set_theta_direction(-1)
        ax6.set_xlabel("$\u03b8_{m}$ ($\degree$)")
        ax6.set_ylabel("$H_{s}$ (m)", labelpad=20)
        ax6.set_title("SWELL 1 $Wave$ $Climate$", pad=35, fontweight="bold")

        return fig, [ax1, ax2, ax1_cbar, ax2_cbar, ax3, ax4, ax5, ax6]

    def validate_calibration(
        self, data_to_validate: pd.DataFrame
    ) -> Tuple[Figure, list]:
        """
        Validate the calibration using independent validation data.

        This method compares the original and corrected model data to the validation data,
        both as time series and with scatter plots.

        Parameters
        ----------
        data_to_validate : pd.DataFrame
            Validation data. Must contain column:
                - 'Hs_VAL' (float): Validation significant wave height (e.g., from buoy)
            The index must be datetime-like.

        Returns
        -------
        Tuple[Figure, list]
            The matplotlib Figure and a list of Axes objects:
            [time series axis, scatter (no correction), scatter (corrected)].
        """

        if "Hs_VAL" not in data_to_validate.columns:
            raise ValueError("Validation data is missing required column: 'Hs_VAL'")

        data_corr = self.correct(data=self._data)
        data_times, data_to_validate_times = get_matching_times_between_arrays(
            times1=data_corr.index,
            times2=data_to_validate.index,
            max_time_diff=1,
        )

        # Create figure with a 2-row, 2-column grid, top row spans both columns
        fig = plt.figure(figsize=(12, 8))
        gs = fig.add_gridspec(2, 2, height_ratios=[2, 3], hspace=0.4, wspace=0.3)

        # Top row: time series plot (spans both columns)
        ax_ts = fig.add_subplot(gs[0, :])
        t = data_corr.index[data_times]
        ax_ts.plot(
            t,
            data_to_validate["Hs_VAL"].iloc[data_to_validate_times],
            label="Validation",
            color="k",
            lw=1.5,
        )
        ax_ts.plot(
            t,
            data_corr["Hs"].iloc[data_times],
            label="Model (No Correction)",
            color="tab:blue",
            alpha=0.7,
        )
        ax_ts.plot(
            t,
            data_corr["Hs_CORR"].iloc[data_times],
            label="Model (Corrected)",
            color="tab:orange",
            alpha=0.7,
        )
        ax_ts.set_ylabel("$H_s$ (m)")
        ax_ts.set_xlabel("Time")
        ax_ts.set_title("Time Series Comparison")
        ax_ts.legend(loc="upper right")
        ax_ts.grid(True, linestyle=":", alpha=0.5)

        # Bottom row: scatter plots
        ax_sc1 = fig.add_subplot(gs[1, 0])
        ax_sc2 = fig.add_subplot(gs[1, 1])
        validation_scatter(
            axs=ax_sc1,
            x=data_corr["Hs"].iloc[data_times].values,
            y=data_to_validate["Hs_VAL"].iloc[data_to_validate_times].values,
            xlabel="Model (No Correction)",
            ylabel="Validation",
            title="No Correction",
        )
        validation_scatter(
            axs=ax_sc2,
            x=data_corr["Hs_CORR"].iloc[data_times].values,
            y=data_to_validate["Hs_VAL"].iloc[data_to_validate_times].values,
            xlabel="Model (Corrected)",
            ylabel="Validation",
            title="With Correction",
        )

        return fig, [ax_ts, ax_sc1, ax_sc2]



import inspect
import os.path as op

import numpy as np

from bluemath_tk.datamining.lhs import LHS
from bluemath_tk.datamining.mda import MDA
from bluemath_tk.wrappers.swash.swash_wrapper import HySwashVeggyModelWrapper


if __name__ == "__main__":
    # Define the output directory
    output_dir = (
        "test_cases/HySwashVeggy/"  # CHANGE THIS TO YOUR DESIRED OUTPUT DIRECTORY!
    )
    # Templates directory
    swash_file_path = op.dirname(inspect.getfile(HySwashVeggyModelWrapper))
    templates_dir = op.join(swash_file_path, "templates")
    # Fixed parameters
    fixed_parameters = {
        "dxinp": 1,  # bathymetry grid spacing
        "Plants_ini": 750,  # Vegetation start cell
        "Plants_fin": 900,  # Vegetation end cell
        "comptime": 7200,  # Simulation duration (s)
        "warmup": 7200 * 0.15,  # Warmup duration (s)
        "n_nodes_per_wavelength": 80,  # number of nodes per wavelength
    }
    # LHS
    variables_to_analyse_in_metamodel = [
        "Hs",
        "Hs_L0",
        "WL",
        "vegetation_height",
        "plants_density",
    ]
    lhs_parameters = {
        "num_dimensions": 5,
        "num_samples": 10000,
        "dimensions_names": variables_to_analyse_in_metamodel,
        "lower_bounds": [0.5, 0.005, 0, 0, 0],
        "upper_bounds": [2, 0.05, 1, 1.5, 1000],
    }
    lhs = LHS(num_dimensions=len(variables_to_analyse_in_metamodel))
    df_dataset = lhs.generate(
        dimensions_names=lhs_parameters.get("dimensions_names"),
        lower_bounds=lhs_parameters.get("lower_bounds"),
        upper_bounds=lhs_parameters.get("upper_bounds"),
        num_samples=lhs_parameters.get("num_samples"),
    )
    # MDA
    mda_parameters = {"num_centers": 5}
    mda = MDA(num_centers=mda_parameters.get("num_centers"))
    mda.fit(data=df_dataset)
    metamodel_parameters = mda.centroids.to_dict(orient="list")
    # HySwashVeggyModelWrapper
    swash_wrapper = HySwashVeggyModelWrapper(
        templates_dir=templates_dir,
        metamodel_parameters=metamodel_parameters,
        fixed_parameters=fixed_parameters,
        output_dir=output_dir,
        depth_array=np.loadtxt(op.join(templates_dir, "depth.bot")),
    )
    # Build the input files
    swash_wrapper.build_cases(mode="one_by_one")
    # Run the simulations
    swash_wrapper.run_cases(launcher="docker_serial", num_workers=5)
    # Post-process the results
    swash_wrapper.postprocess_cases(
        output_vars=["Ru2", "Runlev", "Msetup", "Hrms", "Hfreqs"]
    )
    print("Done!")
